"""
General utility functions for conversion and else
"""
from typing import Dict, Any, List, Union, NamedTuple
from omspy.models import BasicPosition
from collections import defaultdict


class UQty(NamedTuple):
    q: float
    f: float
    p: float
    c: float


def create_basic_positions_from_orders_dict(
    orders: List[Dict],
) -> Dict[str, BasicPosition]:
    """
    Create a dictionary of positions from list of orders received from broker
    orders
        list of orders
    returns a dictionary with key being the symbol and values as a BasicPosition model
    """
    dct: Dict[str, BasicPosition] = {}
    for order in orders:
        symbol = order.get("symbol")
        if symbol:
            price = max(
                order.get("average_price", 0),
                order.get("price", 0),
                order.get("trigger_price", 0),
            )
            quantity = abs(order.get("quantity", 0))
            side = order["side"].lower()
            if symbol not in dct.keys():
                # Create symbol if it doesn't exist
                dct[symbol] = BasicPosition(symbol=symbol)
            position = dct[symbol]
            if side == "buy":
                position.buy_quantity += quantity
                position.buy_value += price * quantity
            elif side == "sell":
                position.sell_quantity += quantity
                position.sell_value += price * quantity
    return dct


def dict_filter(lst: List[Dict], **kwargs) -> List[Dict]:
    """
    Filter a list of dictionary to conditions matching
    in kwargs
    lst
        list of dictionaries
    kwargs
        key values to filter; key is the dictionary key
        and value is the value to match.
        **This is an AND filter**

    Note
    -----
    For each dictionary in the list, each of the arguments
    in kwargs are matched and only those dictionaries that
    match all the conditions are returned
    """
    if len(lst) == 0:
        print("Nothing in the list")
        return []
    new_lst = []
    for d in lst:
        case = True
        for k, v in kwargs.items():
            if d.get(k) != v:
                case = False
        if case:
            new_lst.append(d)
    return new_lst


def tick(price, tick_size=0.05):
    """
    Rounds a given price to the requested tick
    """
    return round(price / tick_size) * tick_size


def stop_loss_step_decimal(
    price: float, side: str = "B", dec: float = 0.45, step: int = 2
) -> float:
    """
    Truncates down the stop loss value to the desired step
    and adds the given decimal
    price
        stop loss price
    side
        side to place order, the actual stop loss side
        B for BUY, S for SELL
    dec
        fixed decimal to be added
    step
        step size to determine stop
    Note
    ----
    1. Step object is always a positive number
    2. Side is the actual stop loss side you are placing the order
    """
    step = abs(step)
    m = int(price / step)
    val = (m + 1.0) * step if side == "S" else (m * step) - 1.0
    val = val + 1 - dec if side == "S" else val + dec
    return val


def update_quantity(q: float, f: float, p: float, c: float) -> UQty:
    """
    Update filled,pending and canceled quantity based on the given data
    q
        actual quantity orders
    f
        filled quantity
    p
        pending quantity
    c
        canceled/rejected quantity
    returns a named tuple containing the calculated values
    Note
    -----
    1) q is taken to be the actual quantity
    2) formula used is **quantity = filled_quantity + pending_quantiy - canceled quantity**
    3) canceled quantity is given preference over other quantities, filled quantity
    the next and pending the last priority
    4) if filled or canceled quantity is greater than given quantity, it is set equal
    to the quantity and all other quantities are discarded
    """
    if c > 0:
        c = min(c, q)
        f = q - c
        p = q - c - f
    elif f > 0:
        f = min(f, q)
        p = q - f
    elif p > 0:
        p = min(p, q)
        f = q - p
    else:
        p = q - p
    return UQty(q, f, p, c)



from typing import Optional, Type
from pydantic import BaseModel
from fastapi import FastAPI
from omspy.simulation.virtual import FakeBroker
from omspy.simulation.models import (
    OrderResponse,
    Side,
    Status,
    AuthResponse,
    LTPResponse,
    OHLCVResponse,
    QuoteResponse,
    OrderBookResponse,
    PositionResponse,
    ResponseStatus,
)

description = """
Generate ready to use fake data to mimic
stock market data for testing broker interface
and in general

### User
Generates data with respect to individual user

### Order
Generates data for placing, modifying and canceling orders

### Market Data
Generates data regarding price, OHLC, orderbook for symbols
"""


class MyAPI(FastAPI):
    """
    Sub-classing FastAPI to include our variables
    """

    broker: FakeBroker = FakeBroker()

    @property
    def _type(self) -> Type[FakeBroker]:
        return type(self.broker)


app: MyAPI = MyAPI(title="Fake Data API for Stock Market", description=description)

SUCCESS = ResponseStatus.SUCCESS
FAILURE = ResponseStatus.FAILURE


class OrderArgs(BaseModel):
    symbol: Optional[str]
    side: Optional[Side]
    quantity: Optional[int]
    price: Optional[float]
    trigger_price: Optional[float]
    s: Optional[Status]


class CreateArgs(BaseModel):
    symbol: str
    side: Side
    quantity: Optional[int]
    price: Optional[float]
    trigger_price: Optional[float]


class ModifyArgs(BaseModel):
    quantity: Optional[float]
    price: Optional[float]
    trigger_price: Optional[float]


@app.get("/", summary="Fake Stock Data")
def home():
    return {"hello": "Welcome to Fake Stock Data"}


@app.post(
    "/auth/{user_id}",
    summary="Authenticate the user",
    description="Authenticate the user",
    tags=["user"],
)
async def auth(user_id: str) -> AuthResponse:
    return AuthResponse(status=SUCCESS, user_id=user_id)


@app.post(
    "/order",
    summary="Create an order",
    description="Create an order with the given arguments",
    tags=["order"],
)
async def create_order(order: OrderArgs) -> OrderResponse:
    if app._type == FakeBroker:
        response = app.broker.order_place(**order.dict(exclude_none=True))
    return OrderResponse(status=SUCCESS, data=response)


@app.put(
    "/order/{order_id}",
    summary="Modify an order",
    description="Modify a existing order with the given order_id",
    tags=["order"],
)
async def modify_order(order_id: str, order: OrderArgs) -> OrderResponse:
    if app._type == FakeBroker:
        response = app.broker.order_modify(
            order_id=order_id, **order.dict(exclude_none=True)
        )
    return OrderResponse(status=SUCCESS, data=response)


@app.delete(
    "/order/{order_id}",
    summary="Delete order",
    description="Delete a existing order with the given order_id",
    tags=["order"],
)
async def cancel_order(order_id: str, order: OrderArgs) -> OrderResponse:
    if app._type == FakeBroker:
        response = app.broker.order_cancel(
            order_id=order_id, **order.dict(exclude_none=True)
        )
    return OrderResponse(status=SUCCESS, data=response)


@app.get(
    "/ltp/{symbol}",
    summary="Get the last price for the given symbol",
    tags=["market data"],
)
async def ltp(symbol: str) -> LTPResponse:
    response = app.broker.ltp(symbol)
    return LTPResponse(status=SUCCESS, data=response)


@app.get(
    "/ohlc/{symbol}",
    summary="Get the ohlc prices and volume for the given symbol",
    tags=["market data"],
)
async def ohlc(symbol: str) -> OHLCVResponse:
    response = app.broker.ohlc(symbol)
    return OHLCVResponse(status=SUCCESS, data=response)


@app.get(
    "/quote/{symbol}",
    summary="Get the detailed quote for the symbol",
    description="Quote contains OHLC, volume, last_price and the orderbook",
    tags=["market data"],
)
async def quote(symbol: str) -> QuoteResponse:
    response = app.broker.quote(symbol)
    return QuoteResponse(status=SUCCESS, data=response)


@app.get(
    "/orderbook/{symbol}",
    summary="Get the orderbook for the given symbol",
    tags=["market data"],
)
async def orderbook(symbol: str) -> OrderBookResponse:
    response = app.broker.orderbook(symbol)
    return OrderBookResponse(status=SUCCESS, data=response)


@app.get("/positions", summary="Get random positions", tags=["user"])
async def positions() -> PositionResponse:
    response = app.broker.positions()
    return PositionResponse(status=SUCCESS, data=response)



from omspy.brokers.api_helper import ShoonyaApiPy
from omspy.base import Broker, pre, post
from typing import Optional, List, Dict, Union, Set
import pendulum
import pyotp
import logging


class Finvasia(Broker):
    """
    Automated Trading class
    """

    def __init__(
        self,
        user_id: str,
        password: str,
        pin: str,
        vendor_code: str,
        app_key: str,
        imei: str,
    ):
        self._user_id = user_id
        self._password = password
        self._pin = pin
        self._vendor_code = vendor_code
        self._app_key = app_key
        self._imei = imei
        self.finvasia = ShoonyaApiPy()
        super(Finvasia, self).__init__()

    @property
    def attribs_to_copy_modify(self) -> Set:
        return {"symbol", "exchange"}

    def login(self) -> Union[Dict, None]:
        return self.finvasia.login(
            userid=self._user_id,
            password=self._password,
            twoFA=pyotp.TOTP(self._pin).now(),
            vendor_code=self._vendor_code,
            api_secret=self._app_key,
            imei=self._imei,
        )

    def authenticate(self) -> Union[Dict, None]:
        """
        Authenticate the user
        """
        return self.login()

    def _convert_symbol(self, symbol: str, exchange: str = "NSE") -> str:
        """
        Convert raw symbol to finvasia
        """
        if exchange == "NSE":
            if symbol.endswith("-EQ") or symbol.endswith("-eq"):
                return symbol
            else:
                return f"{symbol}-EQ"
        else:
            return symbol

    @property
    @post
    def orders(self) -> List[Dict]:
        orderbook = self.finvasia.get_order_book()
        if len(orderbook) == 0:
            return orderbook

        order_list = []
        float_cols = ["avgprc", "prc", "rprc", "trgprc"]
        int_cols = ["fillshares", "qty"]
        for order in orderbook:
            try:
                for int_col in int_cols:
                    order[int_col] = int(order.get(int_col, 0))
                for float_col in float_cols:
                    order[float_col] = float(order.get(float_col, 0))
                ts = order["exch_tm"]
                # Timestamp converted to str to facilitate loading into pandas dataframe
                order["exchange_timestamp"] = str(
                    pendulum.from_format(
                        ts, fmt="DD-MM-YYYY HH:mm:ss", tz="Asia/Kolkata"
                    )
                )
                ts2 = order["norentm"]
                order["broker_timestamp"] = str(
                    pendulum.from_format(
                        ts2, fmt="HH:mm:ss DD-MM-YYYY", tz="Asia/Kolkata"
                    )
                )
            except Exception as e:
                logging.error(e)
            order_list.append(order)
        return order_list

    @property
    @post
    def positions(self) -> List[Dict]:
        positionbook = self.finvasia.get_positions()
        if len(positionbook) == 0:
            return positionbook

        position_list = []
        int_cols = [
            "netqty",
            "daybuyqty",
            "daysellqty",
            "cfbuyqty",
            "cfsellqty",
            "openbuyqty",
            "opensellqty",
        ]
        float_cols = [
            "daybuyamt",
            "daysellamt",
            "lp",
            "rpnl",
            "dayavgprc",
            "daybuyavgprc",
            "daysellavgprc",
            "urmtom",
        ]
        for position in positionbook:
            try:
                for int_col in int_cols:
                    position[int_col] = int(position.get(int_col, 0))
                for float_col in float_cols:
                    position[float_col] = float(position.get(float_col, 0))
            except Exception as e:
                logging.error(e)
            position_list.append(position)
        return position_list

    @property
    @post
    def trades(self) -> List[Dict]:
        tradebook = self.finvasia.get_trade_book()
        if len(tradebook) == 0:
            return tradebook

        trade_list = []
        int_cols = ["flqty", "qty", "fillshares"]
        float_cols = ["prc", "flprc"]
        for trade in tradebook:
            try:
                for int_col in int_cols:
                    trade[int_col] = int(trade.get(int_col, 0))
                for float_col in float_cols:
                    trade[float_col] = float(trade.get(float_col, 0))
            except Exception as e:
                logging.error(e)
            trade_list.append(trade)
        return trade_list

    def get_order_type(self, order_type: str) -> str:
        """
        Convert a generic order type to this specific
        broker's order type string
        returns MKT if the order_type is not matching
        """
        order_types = dict(
            LIMIT="LMT", MARKET="MKT", SL="SL-LMT", SLM="SL-MKT", SLL="SL-LMT"
        )
        order_types["SL-M"] = "SL-MKT"
        order_types["SL-L"] = "SL-LMT"
        return order_types.get(order_type.upper(), "MKT")

    @pre
    def order_place(self, **kwargs) -> Union[str, None]:
        side = kwargs.pop("side")
        order_type = kwargs.pop("order_type", "MKT")
        exchange = kwargs.pop("exchange", "NSE")
        symbol = kwargs.pop("symbol")
        symbol = self._convert_symbol(symbol, exchange=exchange)
        if order_type:
            order_type = self.get_order_type(order_type)
        if side:
            side = side.upper()[0]
        if symbol:
            symbol = symbol.upper()
        order_args = dict(
            tradingsymbol=symbol,
            buy_or_sell=side,
            price_type=order_type,
            exchange=exchange,
            retention="DAY",
            product_type="I",
            discloseqty=0,
        )
        order_args.update(kwargs)
        response = self.finvasia.place_order(**order_args)
        return response.get("norenordno")

    def order_cancel(self, order_id: str) -> Union[Dict, None]:
        """
        Cancel an existing order
        """
        return self.finvasia.cancel_order(orderno=order_id)

    @pre
    def order_modify(self, **kwargs) -> Union[str, None]:
        """
        Modify an existing order
        """
        symbol = kwargs.pop("tradingsymbol")
        order_id = kwargs.pop("order_id", None)
        order_type = kwargs.pop("order_type", "MKT")
        if "discloseqty" in kwargs:
            kwargs.pop("discloseqty")
        if order_type:
            order_type = self.get_order_type(order_type)
        if symbol:
            symbol = self._convert_symbol(symbol).upper()
        order_args = dict(
            orderno=order_id,
            newprice_type=order_type,
            exchange="NSE",
            tradingsymbol=symbol,
        )
        order_args.update(kwargs)
        return self.finvasia.modify_order(**order_args)



from NorenRestApiPy.NorenApi import NorenApi
from threading import Timer
import pandas as pd
import time
import concurrent.futures

api = None


class Order:
    def __init__(
        self,
        buy_or_sell: str = None,
        product_type: str = None,
        exchange: str = None,
        tradingsymbol: str = None,
        price_type: str = None,
        quantity: int = None,
        price: float = None,
        trigger_price: float = None,
        discloseqty: int = 0,
        retention: str = "DAY",
        remarks: str = "tag",
        order_id: str = None,
    ):
        self.buy_or_sell = buy_or_sell
        self.product_type = product_type
        self.exchange = exchange
        self.tradingsymbol = tradingsymbol
        self.quantity = quantity
        self.discloseqty = discloseqty
        self.price_type = price_type
        self.price = price
        self.trigger_price = trigger_price
        self.retention = retention
        self.remarks = remarks
        self.order_id = None


# print(ret)


def get_time(time_string):
    data = time.strptime(time_string, "%d-%m-%Y %H:%M:%S")

    return time.mktime(data)


class ShoonyaApiPy(NorenApi):
    def __init__(self):
        NorenApi.__init__(
            self,
            host="https://api.shoonya.com/NorenWClientTP/",
            websocket="wss://api.shoonya.com/NorenWSTP/",
            eodhost="https://api.shoonya.com/chartApi/getdata/",
        )
        global api
        api = self

    def place_basket(self, orders):

        resp_err = 0
        resp_ok = 0
        result = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:

            future_to_url = {
                executor.submit(self.place_order, order): order for order in orders
            }
            for future in concurrent.futures.as_completed(future_to_url):
                url = future_to_url[future]
            try:
                result.append(future.result())
            except Exception as exc:
                print(exc)
                resp_err = resp_err + 1
            else:
                resp_ok = resp_ok + 1

        return result

    def placeOrder(self, order: Order):
        ret = NorenApi.place_order(
            self,
            buy_or_sell=order.buy_or_sell,
            product_type=order.product_type,
            exchange=order.exchange,
            tradingsymbol=order.tradingsymbol,
            quantity=order.quantity,
            discloseqty=order.discloseqty,
            price_type=order.price_type,
            price=order.price,
            trigger_price=order.trigger_price,
            retention=order.retention,
            remarks=order.remarks,
        )
        # print(ret)

        return ret



"""
This module contains the list of basic models
"""
from pydantic import BaseModel, validator, PrivateAttr
from typing import Optional, List, Union
from copy import deepcopy
import pendulum
import logging


class QuantityMatch(BaseModel):
    buy: int = 0
    sell: int = 0

    @property
    def is_equal(self) -> bool:
        return True if self.buy == self.sell else False

    @property
    def not_matched(self) -> int:
        return self.buy - self.sell


class BasicPosition(BaseModel):
    """
    A simple position tracking mechanism
    """

    symbol: str
    buy_quantity: int = 0
    sell_quantity: int = 0
    buy_value: float = 0.0
    sell_value: float = 0.0

    @property
    def net_quantity(self) -> int:
        return self.buy_quantity - self.sell_quantity

    @property
    def average_buy_value(self) -> float:
        return self.buy_value / self.buy_quantity if self.buy_value > 0.0 else 0.0

    @property
    def average_sell_value(self) -> float:
        return self.sell_value / self.sell_quantity if self.sell_quantity > 0 else 0.0


class Quote(BaseModel):
    price: float
    quantity: int
    orders_count: Optional[int] = None

    @property
    def value(self):
        return self.price * self.quantity


class OrderBook(BaseModel):
    bid: List[Quote]
    ask: List[Quote]

    @property
    def is_bid_ask(self) -> bool:
        """
        return True if there is atleast one bid and ask else False
        """
        if len(self.bid) > 0 and len(self.ask) > 0:
            return True
        else:
            return False

    @property
    def spread(self) -> float:
        """
        returns the bid-ask spread, 0 if not available
        """
        if self.is_bid_ask:
            return self.ask[0].price - self.bid[0].price
        else:
            return 0.0

    @property
    def total_bid_quantity(self) -> int:
        """
        returns the total bid quantity
        """
        if self.is_bid_ask:
            return sum((x.quantity for x in self.bid))
        else:
            return 0

    @property
    def total_ask_quantity(self) -> int:
        """
        returns the total ask quantity
        """
        if self.is_bid_ask:
            return sum((x.quantity for x in self.ask))
        else:
            return 0


class Tracker(BaseModel):
    """
    A simple tracker to track the high low prices
    """

    name: str  # name of the symbol to be tracked
    last_price: float = 0
    # Initializing high and low to extreme values
    high: float = -1e100
    low: float = 1e100

    def update(self, last_price: float):
        self.last_price = last_price
        self.high = max(last_price, self.high)
        self.low = min(last_price, self.low)


class Timer(BaseModel):
    """
    A simple timer that could be attached to any model
    """

    start_time: pendulum.DateTime
    end_time: pendulum.DateTime
    timezone: Optional[str] = None

    @validator("end_time")
    def validate_times(cls, v, values):
        """
        Validate end_time greater than start_time
        and start_time greater than current time
        """
        start = values.get("start_time")
        tz = values.get("timezone")
        if v < start:
            raise ValueError("end time greater than start time")
        if start < pendulum.now(tz=tz):
            raise ValueError("start time lesser than current time")
        return v

    @property
    def has_started(self) -> bool:
        """
        Whether tracking has started
        """
        return True if pendulum.now(tz=self.timezone) > self.start_time else False

    @property
    def has_completed(self) -> bool:
        """
        Whether tracking has completed
        """
        return True if pendulum.now(tz=self.timezone) > self.end_time else False

    @property
    def is_running(self) -> bool:
        """
        Whether the timer is in progress
        returns True if it has started and not completed
        else False
        """
        if self.has_started and not self.has_completed:
            return True
        else:
            return False


class TimeTracker(Tracker, Timer):
    pass


class OrderLock(BaseModel):
    """
    Lock order placement, modification and cancellation
    for a few seconds
    """

    max_order_creation_lock_time: float = 60
    max_order_modification_lock_time: float = 60
    max_order_cancellation_lock_time: float = 60
    timezone: Optional[str] = None
    _creation_lock_till: pendulum.DateTime = PrivateAttr()
    _modification_lock_till: pendulum.DateTime = PrivateAttr()
    _cancellation_lock_till: pendulum.DateTime = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._creation_lock_till = pendulum.now(tz=self.timezone)
        self._modification_lock_till = pendulum.now(tz=self.timezone)
        self._cancellation_lock_till = pendulum.now(tz=self.timezone)

    @property
    def creation_lock_till(self):
        return self._creation_lock_till

    @property
    def modification_lock_till(self):
        return self._modification_lock_till

    @property
    def cancellation_lock_till(self):
        return self._cancellation_lock_till

    def create(self, seconds: float) -> pendulum.DateTime:
        """
        Lock the create_order function for the given number of seconds
        """
        seconds = min(seconds, self.max_order_creation_lock_time)
        self._creation_lock_till = pendulum.now(tz=self.timezone).add(
            seconds=int(seconds)
        )
        return self.creation_lock_till

    def modify(self, seconds: float) -> pendulum.DateTime:
        """
        Lock the modify_order function for the given number of seconds
        """
        seconds = min(seconds, self.max_order_modification_lock_time)
        self._modification_lock_till = pendulum.now(tz=self.timezone).add(
            seconds=int(seconds)
        )
        return self.modification_lock_till

    def cancel(self, seconds: float) -> pendulum.DateTime:
        """
        Lock the cancel_order function for the given number of seconds
        """
        seconds = min(seconds, self.max_order_cancellation_lock_time)
        self._cancellation_lock_till = pendulum.now(tz=self.timezone).add(
            seconds=int(seconds)
        )
        return self.cancellation_lock_till

    @property
    def can_create(self) -> bool:
        """
        returns True if order can be created else False
        """
        return (
            True if pendulum.now(tz=self.timezone) > self.creation_lock_till else False
        )

    @property
    def can_modify(self) -> bool:
        """
        returns True if order can be modified else False
        """
        return (
            True
            if pendulum.now(tz=self.timezone) > self.modification_lock_till
            else False
        )

    @property
    def can_cancel(self) -> bool:
        """
        returns True is order can be canceled else False
        """
        return (
            True
            if pendulum.now(tz=self.timezone) > self.cancellation_lock_till
            else False
        )


class Candle(BaseModel):
    """
    A model representing a single candle
    """

    timestamp: pendulum.DateTime
    open: float
    high: float
    low: float
    close: float
    volume: Optional[float] = None
    info: Optional[str] = None


class CandleStick(BaseModel):
    """
    Class to work on candlesticks
    """

    symbol: str
    candles: List[Candle] = []
    initial_price: float = 0
    interval: int = 300  # in seconds
    timer: Optional[Timer]
    timezone: str = "local"
    ltp: float = 0
    high: float = -1e100  # Initialize to a impossible value
    low: float = 1e100  # Initialize to a impossible value
    bar_open: float = 0
    bar_high: float = -1e100  # Initialize to a impossible value
    bar_low: float = 1e100  # Initialize to a impossible value
    next_interval: Optional[pendulum.DateTime] = None
    periods: List[pendulum.DateTime] = []
    _last_ltp: float = PrivateAttr()  # to track bar close price

    class Config:
        underscore_attribs_are_private = True

    def __init__(self, **data) -> None:
        super().__init__(**data)
        self._last_ltp = 0
        if self.timer is None:
            timer = Timer(
                start_time=pendulum.today(tz=self.timezone).add(hours=9, minutes=15),
                end_time=pendulum.today(tz=self.timezone).add(hours=15, minutes=30),
            )
            self.timer = timer
        period = pendulum.period(self.timer.start_time, self.timer.end_time)
        for p in period.range("seconds", self.interval):
            self.periods.append(p)
        # The first period is popped since it is the start
        self.periods.pop(0)
        if self.next_interval is None:
            self.next_interval = self.periods.pop(0)

    def add_candle(self, candle: Candle) -> None:
        """
        Add a candle
        """
        self.candles.append(deepcopy(candle))

    def _update_prices(self):
        """
        Update running candle
        """
        ltp = self.ltp
        if self.initial_price == 0:
            self.initial_price = ltp
        if self.bar_open == 0:
            self.bar_open = ltp
        self.bar_high = max(self.bar_high, ltp)
        self.bar_low = min(self.bar_low, ltp)
        self.high = max(self.high, ltp)
        self.low = min(self.low, ltp)

    def update_candle(self, timestamp: pendulum.DateTime = pendulum.now()) -> Candle:
        """
        Update and append the existing candle
        returns the updated candle
        """
        if len(self.candles) == 0:
            open_price = self.initial_price
        else:
            open_price = self.bar_open
        candle = Candle(
            timestamp=timestamp,
            open=open_price,
            high=self.bar_high,
            low=self.bar_low,
            close=self._last_ltp,
        )
        self.add_candle(candle)
        self.bar_high = -1e100
        self.bar_low = 1e100
        self.bar_open = 0
        self._update_prices()
        return candle

    @property
    def bullish_bars(self) -> int:
        """
        Returns the number of bullish bars
        """
        count = 0
        for candle in self.candles:
            if candle.close > candle.open:
                count += 1
        return count

    @property
    def bearish_bars(self) -> int:
        """
        Returns the number of bullish bars
        """
        count = 0
        for candle in self.candles:
            if candle.close < candle.open:
                count += 1
        return count

    def get_next_interval(self) -> Union[pendulum.DateTime, None]:
        """
        Get the next time interval
        returns None if all intervals are completed
        """
        if len(self.periods) == 0:
            return None
        to_remove = []
        period = None
        for p in self.periods:
            if p > pendulum.now(tz=self.timezone):
                period = p
                to_remove.append(p)
                break
            else:
                to_remove.append(p)
        for p in to_remove:
            try:
                self.periods.remove(p)
            except ValueError:
                logging.error(f"Period {period} cannot be found in the list of periods")
        return period

    def update(self, ltp: float) -> None:
        if self.timer and self.next_interval:
            if self.timer.is_running:
                self._last_ltp = self.ltp
                self.ltp = ltp
                now = pendulum.now(tz=self.timezone)
                if now > self.next_interval:
                    self.update_candle(timestamp=self.next_interval)
                    self.next_interval = self.get_next_interval()
                else:
                    self._update_prices()

    @property
    def last_bullish_bar_index(self) -> int:
        """
        Get the index of the last bullish bar candle
        """
        l = len(self.candles)
        if l == 0:
            return 0
        for i, candle in enumerate(reversed(self.candles)):
            if candle.close > candle.open:
                return l - i
        # If no bullish candle
        return 0

    @property
    def last_bearish_bar_index(self) -> int:
        """
        Get the index of the last bullish bar candle
        """
        l = len(self.candles)
        if l == 0:
            return 0
        for i, candle in enumerate(reversed(self.candles)):
            if candle.close < candle.open:
                return l - i
        # If no bearish candle
        return 0

    @property
    def last_bullish_bar(self) -> Optional[Candle]:
        """
        Return the latest bullish bar
        """
        l = len(self.candles)
        if l == 0:
            return None
        i = self.last_bullish_bar_index
        if i == 0:
            return None
        else:
            return self.candles[i - 1]

    @property
    def last_bearish_bar(self) -> Optional[Candle]:
        """
        Return the latest bearish bar
        """
        l = len(self.candles)
        if l == 0:
            return None
        i = self.last_bearish_bar_index
        if i == 0:
            return None
        else:
            return self.candles[i - 1]



from omspy.base import Broker, post, pre
from typing import Optional, List, Dict, Union
from neo_api_client import NeoAPI
import pendulum
import logging


class Neo(Broker):
    """
    Automated trading class for Neo Broker
    """

    def __init__(
        self,
        consumer_key: str,
        consumer_secret: str,
        mobilenumber: str,
        password: str,
        twofa: str,
        user_id: Optional[str] = None,
        **kwargs,
    ):
        self._user_id = user_id
        self._mobilenumber = mobilenumber
        self._password = password
        self._consumer_key = consumer_key
        self._consumer_secret = consumer_secret
        self._mpin = twofa
        self._kwargs = kwargs
        super(Neo, self).__init__()
        client = NeoAPI(
            consumer_key=self._consumer_key,
            consumer_secret=self._consumer_secret,
            **self._kwargs,
        )
        self.neo = client

    def authenticate(self) -> Dict:
        self.neo.login(
            password=self._password,
            mobilenumber=self._mobilenumber,
        )
        return self.neo.session_2fa(self._mpin)

    @pre
    def order_place(self, **kwargs) -> Union[str, None]:
        """
        place an order
        """
        try:
            order_args = dict(
                exchange_segment="NSE",
                product="MIS",
                order_type="MKT",
                validity="DAY",
            )
            order_args["transaction_type"] = kwargs.pop("transaction_type").upper()[0]
            for key in ("quantity", "price", "trigger_price", "disclosed_quantity"):
                val = str(kwargs.pop(key, 0))
                order_args.update({key: val})
            order_args.update(kwargs)
            response = self.neo.place_order(**order_args)
            if response.get("Error"):
                logging.error(response["Error"])
                return None
            elif response.get("error"):
                logging.error(response["error"])
                return None
            else:
                return response.get("nOrdNo")
        except Exception as e:
            logging.error(e)
            return None

    @pre
    def order_modify(self, order_id: str, **kwargs) -> Optional[Dict]:
        """
        modify the order
        """
        modify_args = dict(validity="DAY", product="MIS", amo="NO")
        for key in ("quantity", "price", "trigger_price", "disclosed_quantity"):
            if key in kwargs:
                kwargs[key] = str(kwargs[key])
        modify_args.update(kwargs)
        response = self.neo.modify_order(order_id=order_id, **modify_args)
        return response

    def order_cancel(self, order_id: str):
        """
        cancel an existing order
        """
        response = self.neo.cancel_order(order_id=order_id)
        return response

    @property
    @post
    def orders(self) -> List[Dict]:
        """
        return the list of orders
        """
        int_cols = ["cnlQty", "qty", "dscQty", "fldQty"]
        float_cols = ["prc", "trgPrc", "avgPrc", "refLmtPrc"]
        response = self.neo.order_report()
        if "data" in response:
            orderbook = response["data"]
            for o in orderbook:
                try:
                    o["ordSt"] = str(o["ordSt"]).upper()
                    for col in int_cols:
                        if col in o:
                            o[col] = int(o[col])
                    for col in float_cols:
                        if col in o:
                            o[col] = float(o[col])
                except Exception as e:
                    logging.error(e)
            return orderbook
        else:
            logging.warning(response)
            return [{}]

    @property
    @post
    def positions(self) -> List[Dict]:
        """
        return the list of positions
        """
        int_cols = ["cfBuyQty", "cfSellQty", "flBuyQty", "flSellQty"]
        float_cols = ["buyAmt", "cfSellAmt", "cfBuyAmt", "sellAmt"]
        response = self.neo.positions()
        if "data" in response:
            position_book = response["data"]
            for p in position_book:
                try:
                    quantity = int(p["flBuyQty"]) - int(p["flSellQty"])
                    p["quantity"] = quantity
                    if quantity > 0:
                        p["side"] = "BUY"
                    else:
                        p["side"] = "SELL"

                    for col in int_cols:
                        if col in p:
                            p[col] = int(p[col])
                    for col in float_cols:
                        if col in p:
                            p[col] = float(p[col])
                except Exception as e:
                    logging.error(e)
            return position_book
        else:
            return [{}]

    @property
    @post
    def trades(self) -> List[Dict]:
        """
        return the list of trades
        """
        int_cols = ["fldQty"]
        float_cols = ["avgPrc"]
        response = self.neo.trade_report()
        if "data" in response:
            tradebook = response["data"]
            for t in tradebook:
                try:
                    for col in int_cols:
                        if col in t:
                            t[col] = int(t[col])
                    for col in float_cols:
                        if col in t:
                            t[col] = float(t[col])
                except Exception as e:
                    logging.error(e)
            return tradebook
        else:
            return [{}]



import warnings

DEPRECATION_WARNING = """
This Fyers broker module would be removed from version 0.16.0
"""
warnings.warn(DEPRECATION_WARNING, DeprecationWarning, stacklevel=2)

from omspy.base import Broker, pre, post
from typing import Optional, List, Dict
from urllib.parse import urlparse, parse_qs
from fyers_api import fyersModel, accessToken
from copy import deepcopy

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC

# Constant dictionaries
EXCHANGES = {10: "NSE", 11: "MCX", 12: "BSE"}
SEGMENTS = {10: "capital", 11: "equity", 12: "currency", 20: "commodity"}
ORDER_TYPES = {1: "LIMIT", 2: "MARKET", 3: "SL-M", 4: "SL"}
STATUS = {1: "CANCELED", 2: "COMPLETE", 4: "PENDING", 5: "REJECTED", 6: "PENDING"}
SIDES = {1: "buy", -1: "sell"}


def get_key(url, key="request_token") -> Optional[str]:
    """
    Get the required key from the query parameter
    """
    req = urlparse(url)
    key = parse_qs(req.query).get(key)
    if key is None:
        return None
    else:
        return key[0]


class Fyers(Broker):
    def __init__(self, app_id, secret, user_id, password, pan, log_path="."):
        self._app_id = app_id
        self._secret = secret
        self._user_id = user_id
        self._password = password
        self._pan = pan
        self._store_access_token = True
        self._log_path = log_path
        super(Fyers, self).__init__()

    def authenticate(self) -> None:

        """
        Authenticates a fyers session if access token is already available
        Looks up token in fyers_token.tok file
        Useful for reconnecting instead of logging in again
        """
        try:
            with open("fyers_token.tok") as f:
                access_token = f.read()
            self.fyers = fyersModel.FyersModel(
                client_id=self._app_id, token=access_token, log_path=self._log_path
            )
            self.quote = self.fyers.quotes
        except:
            print("Unknown Exception")
            self._login()

    def _login(self) -> None:
        session = accessToken.SessionModel(
            client_id=self._app_id,
            secret_key=self._secret,
            redirect_uri="http://127.0.0.1",
            response_type="code",
            grant_type="authorization_code",
            state="private",
        )
        auth_url = session.generate_authcode()
        options = webdriver.ChromeOptions()
        options.add_argument("--disable-notifications")
        options.add_argument("--headless")
        driver = webdriver.Chrome(options=options)
        driver.get(auth_url)
        driver.find_element_by_id("fyers_id").send_keys(self._user_id)
        driver.find_element_by_id("password").send_keys(self._password)
        driver.find_element_by_id("pancard").send_keys(self._pan)

        driver.find_element_by_xpath("//button[@id='btn_id']").click()
        WebDriverWait(driver, 20).until((EC.url_changes(driver.current_url)))

        parsed = urlparse(driver.current_url)
        auth_code = parse_qs(parsed.query)["auth_code"][0]
        session.set_token(auth_code)
        response = session.generate_token()
        token = response["access_token"]
        with open("fyers_token.tok", "w") as f:
            f.write(token)
        self.fyers = fyersModel.FyersModel(client_id=self._app_id, token=token)
        self.quote = self.fyers.quotes
        driver.close()

    @property
    def profile(self) -> Dict:
        return self.fyers.get_profile()

    @property
    def funds(self) -> Dict:
        return self.fyers.funds()

    @property
    @post
    def orders(self) -> List[Dict]:
        orderbook = self.fyers.orderbook().get("orderBook")
        orderbook = deepcopy(orderbook)
        if orderbook:
            for order in orderbook:
                order["exchange"] = EXCHANGES.get(order["exchange"])
                order["segment"] = SEGMENTS.get(order["segment"])
                order["side"] = SIDES.get(order["side"])
                order["status"] = STATUS.get(order["status"])
                order["type"] = ORDER_TYPES.get(order["type"])

            return orderbook
        else:
            return [{}]

    @property
    @post
    def positions(self) -> List[Dict]:
        position_book = self.fyers.positions().get("netPositions")
        position_book = deepcopy(position_book)
        if position_book:
            for position in position_book:
                position["side"] = SIDES.get(position["side"])
            return position_book
        else:
            return [{}]

    @property
    @post
    def trades(self) -> List[Dict]:
        tradebook = self.fyers.tradebook().get("tradeBook")
        tradebook = deepcopy(tradebook)
        if tradebook:
            for trade in tradebook:
                trade["side"] = SIDES.get(trade["side"])
                trade["exchange"] = EXCHANGES.get(trade["exchange"])
                trade["segment"] = SEGMENTS.get(trade["segment"])
            return tradebook
        else:
            return [{}]

    @pre
    def order_place(
        self,
        symbol: str,
        side: str,
        quantity: int = 1,
        order_type: str = "MARKET",
        **kwargs
    ) -> Dict:
        """
        Place an actual order with broker
        """
        # Reverse look up maps
        rev_sides = {v: k for k, v in SIDES.items()}
        rev_order_types = {v: k for k, v in ORDER_TYPES.items()}
        side = rev_sides.get(side.lower())
        order_type = kwargs.pop("type", "market").upper()
        order_type = rev_order_types.get(order_type)
        if "qty" not in kwargs:
            kwargs["qty"] = 1
        order_args = {"side": side, "type": order_type, "symbol": symbol}
        for k, v in kwargs.items():
            if k not in order_args:
                order_args[k] = v
        response = self.fyers.place_order(order_args)
        return response

    @pre
    def order_modify(self, **kwargs) -> Dict:
        """
        Modify an actual order with the broker
        """
        # Reverse look up map
        rev_order_types = {v: k for k, v in ORDER_TYPES.items()}
        if kwargs.get("type"):
            order_type = kwargs.get("type", "market").upper()
            order_type = rev_order_types.get(order_type)
            kwargs.update({"type": order_type})
        response = self.fyers.modify_order(kwargs)
        return response

    def order_cancel(self, order_id: str) -> Dict:
        """
        Cancel an actual order with the broker
        """
        response = self.fyers.cancel_order({"id": order_id})
        return response



"""
Module for multi-user multi-broker implementation
"""
from pydantic import BaseModel
from omspy.base import Broker
from omspy.order import Order
from typing import Dict, List, Optional, Type
from collections import defaultdict
import logging
import uuid


class User(BaseModel):
    """
    A basic user class for multi user environment
    """

    broker: Broker
    scale: float = 1.0
    name: Optional[str]
    client_id: Optional[str]
    exclude: Optional[Dict]

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True


class UserOrder(BaseModel):
    order: Order
    user: User

    class Config:
        arbitrary_types_allowed = True


class MultiUser:
    """
    Multi-user implementation
    """

    def __init__(self, users: List[User]):
        self._users: List[User] = users
        self._orders: defaultdict(list) = {}

    def add(self, user: User):
        """
        Add a user
        """
        self._users.append(user)

    @property
    def users(self) -> List[User]:
        return self._users

    @property
    def orders(self) -> Dict[str, Order]:
        return self._orders

    @property
    def count(self) -> int:
        return len(self.users)


class MultiOrder(Order):
    _orders: List[UserOrder] = []

    def __init__(self, **data) -> None:
        super().__init__(**data)
        self.pseudo_id = uuid.uuid4().hex
        self.is_multi = True

    @property
    def orders(self) -> List[UserOrder]:
        return self._orders

    @property
    def count(self) -> int:
        """
        Return the number of orders
        """
        return len(self.orders)

    def create(self, users: Optional[MultiUser]) -> List[UserOrder]:
        # Clear existing list
        self._orders.clear()
        for user in users.users:
            order2 = self.clone()
            order2.quantity = int(user.scale * self.quantity)
            order2.pseudo_id = self.pseudo_id
            order2.save_to_db()
            m_order = UserOrder(order=order2, user=user)
            self._orders.append(m_order)
        self.save_to_db()
        return self.orders

    def save_to_db(self) -> bool:
        """
        save or update the order to db
        """
        if self.connection:
            values = [self.dict(exclude=self._exclude_fields)]
            for order in self.orders:
                values.append(order.order.dict(exclude=self._exclude_fields))
            self.connection["orders"].upsert_all(values, pk="id")
            return True
        else:
            logging.info("No valid database connection")
            return False

    def execute(self, broker: MultiUser, **kwargs):
        """
        Execute order on all users
        broker
            A Multi User instance
            name is retained as broker so it is compatible
            with the original Order interface
        """
        if self.count == 0:
            self.create(users=broker)
        for order in self.orders:
            order.order.execute(order.user.broker, **kwargs)

    def modify(self, **kwargs):
        """
        modify all orders
        """
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
        if "quantity" in kwargs:
            kwargs.pop("quantity")
        for order in self.orders:
            quantity = int(self.quantity * order.user.scale)
            order.order.quantity = quantity
            order.order.modify(order.user.broker, quantity=quantity, **kwargs)

    def cancel(self, **kwargs):
        """
        cancel all existing orders
        """
        for order in self.orders:
            order.order.cancel(order.user.broker)

    def update(self, data: Dict[str, Dict]) -> None:
        """
        Update order based on information received from broker
        data
            data to update as dictionary; key should be the broker order_id
        returns True if update is done
        """
        keys = data.keys()
        for order in self._orders:
            order_id = order.order.order_id
            order_details = data.get(order_id)
            if order_details:
                order.order.update(order_details, save=False)
        self.save_to_db()



from pydantic import BaseModel, ValidationError, validator
from typing import List, Optional, Type, Tuple, Union, Dict
from omspy.base import Broker
from omspy.order import Order, CompoundOrder
from omspy.models import Timer
from omspy.orders.peg import PegExisting, PegSequential
import pendulum
import sqlite_utils
import logging


"""
This module is not functional
"""


class BaseStrategy(BaseModel):
    start_time: pendulum.DateTime
    end_time: pendulum.DateTime
    cycle: int = 0
    done: bool = False
    broker: Optional[Type[Broker]] = None
    connection: Optional[sqlite_utils.Database] = None
    timezone: Optional[
        str
    ] = "local"  # "local" is same as None - when used with Timer (make intention explicit)
    _timer: Optional[
        Timer
    ] = None  # BaseStrategy __init__ uses timezone above in __init__ call of Timer
    _pegs: List[Union[PegExisting, PegSequential]] = []

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True

    def __init__(self, **data):
        super().__init__(**data)
        self._timer = Timer(
            start_time=self.start_time, end_time=self.end_time, timezone=self.timezone
        )

    @property
    def timer(self):
        return self._timer

    def update_orders(self, orders: Optional[Dict[str, Dict]] = None):
        """
        Update existing orders
        """
        if orders is None:
            orders = self.broker.orders
        if hasattr(self, "_order"):
            for order in self._order.orders:
                o = orders.get(order.id)
                if o:
                    order.update(o)


class ShortStraddle(BaseStrategy):
    symbols: Tuple[str, str]
    limit_price: Optional[Tuple[float, float]] = (0.0, 0.0)
    trigger_price: Optional[Tuple[float, float]] = (0.0, 0.0)
    stop_price: Optional[Tuple[float, float]] = (0.0, 0.0)
    quantity: int = 1
    exclude_stop: bool = False
    ltp: Dict[str, float] = {}
    _order: Optional[CompoundOrder] = None
    _order_map: Optional[Dict[str, Order]] = None

    def __init__(self, **data):
        super().__init__(**data)
        self._order = CompoundOrder(
            broker=self.broker, connection=self.connection, timezone=self.timezone
        )
        self._order_map = dict(entry1=None, exit1=None, entry2=None, exit2=None)
        for symbol in self.symbols:
            self.ltp[symbol] = 0

    @property
    def order(self) -> CompoundOrder:
        return self._order

    def get_order(self, name: str) -> Union[Order, None]:
        """
        Get order by name/key
        Note
        ----
        1) Orders are named entry1, exit1, entry2, exit2 for easy reference
        """
        return self._order_map.get(name)

    def create_order(self):
        if len(self._order.orders) > 0:
            return
        com = self._order
        s1, s2 = self.symbols
        order1 = Order(symbol=s1, side="sell", quantity=self.quantity)
        order1.price = self.limit_price[0]
        order2 = order1.clone()
        order2.symbol = s2
        order2.price = self.limit_price[1]
        com.add(order1)
        com.add(order2)
        order1stop = order1.clone()
        order1stop.trigger_price = self.trigger_price[0]
        order1stop.price = self.stop_price[0]
        order1stop.order_type = "SL"
        order1stop.side = "buy"
        com.add(order1stop)
        order2stop = order2.clone()
        order2stop.trigger_price = self.trigger_price[1]
        order2stop.price = self.stop_price[1]
        order2stop.order_type = "SL"
        order2stop.side = "buy"
        com.add(order2stop)
        self._order_map["entry1"] = order1
        self._order_map["exit1"] = order1stop
        self._order_map["entry2"] = order2
        self._order_map["exit2"] = order2stop
        return self.order

    @staticmethod
    def _check_orders_complete(one: Order, two: Order) -> bool:
        """
        returns whether the first leg of the order is complete
        Note
        ----
        1) returns True if both orders are complete or rejected or canceled
        2) returns True even if one order is rejected and other order is canceled
        """
        statuses = ("REJECTED", "CANCELED", "CANCELLED")
        if one.is_complete and two.is_complete:
            return True
        elif one.status in statuses and two.status in statuses:
            return True
        else:
            return False

    @property
    def is_first_leg_complete(self) -> bool:
        """
        returns whether the first leg of the order is complete
        Note
        ----
        1) returns True if both orders are complete or rejected or canceled
        2) returns True even if one order is rejected and other order is canceled
        """
        one = self.get_order("entry1")
        two = self.get_order("exit1")
        if one is None or two is None:
            return False
        else:
            return self._check_orders_complete(one, two)

    @property
    def is_second_leg_complete(self) -> bool:
        """
        returns whether the first leg of the order is complete
        Note
        ----
        1) returns True if both orders are complete or rejected or canceled
        2) returns True even if one order is rejected and other order is canceled
        """
        one = self.get_order("entry2")
        two = self.get_order("exit2")
        if one is None or two is None:
            return False
        else:
            return self._check_orders_complete(one, two)

    def _check_sell_without_buy(self, one: Order, two: Order) -> bool:
        """
        one
            the sell order
        two
            the corresponding buy order
        returns True if a sell order is without a buy order
        Note
        ----
        If this returns True, then it is not valid
        """
        status_one = one.status
        status_two = two.status
        statuses = ("REJECTED", "CANCELED", "CANCELLED")
        if one.is_complete and status_two in statuses:
            return True
        elif one.is_pending and status_two in statuses:
            return True
        else:
            return False

    def _check_buy_without_sell(self, one: Order, two: Order) -> bool:
        """
        one
            the sell order
        two
            the corresponding buy order
        returns True if a buy order is without a sell order
        Note
        ----
        If this returns True, then it is not valid
        """
        # Just swapping the arguments
        return self._check_sell_without_buy(two, one)

    def update_ltp(self, ltp: Dict[str, float]):
        """
        Update ltp for the given symbols
        """
        num = len(self.symbols)
        count = 0
        for symbol, last_price in ltp.items():
            if symbol in self.ltp:
                self.ltp[symbol] = last_price
                count += 1
            if count >= num:
                break
        return self.ltp

    def _make_sequential_orders(self):
        """
        Make sequential peg orders
        """
        if len(self._pegs) == 0 and len(self.order.orders) == 4:
            seq1 = PegSequential(
                broker=self.broker,
                timezone=self.timezone,
                start_time=self.start_time,
                orders=[self.get_order("entry1"), self.get_order("exit1")],
            )
            seq2 = PegSequential(
                broker=self.broker,
                timezone=self.timezone,
                start_time=self.start_time,
                orders=[self.get_order("entry2"), self.get_order("exit2")],
            )
            self._pegs.append(seq1)
            self._pegs.append(seq2)



"""
This module contains all the models for running the simulation
All the models start with **V** to indicate virtual models
"""

from pydantic import BaseModel, Field, validator, PrivateAttr
from typing import Optional, Union, Any, Dict, List
from enum import Enum
import random
import pendulum
import omspy.utils as utils
from omspy.models import OrderBook


class Status(Enum):
    COMPLETE = 1
    REJECTED = 2
    CANCELED = 3
    PARTIAL_FILL = 4  # partially filled but completed order
    OPEN = 5  # all quantity is pending to be filled
    PENDING = 6  # partially filled, waiting to get complete


class ResponseStatus(str, Enum):
    SUCCESS = "success"
    FAILURE = "failure"


class Side(Enum):
    BUY = 1
    SELL = -1


class TickerMode(Enum):
    RANDOM = 1
    MANUAL = 2


class OHLC(BaseModel):
    open: float
    high: float
    low: float
    close: float
    last_price: float


class OHLCV(OHLC):
    volume: int


class OHLCVI(OHLCV):
    open_interest: int


class Ticker(BaseModel):
    """
    A simple ticker class to generate fake data
    name
        name for this ticker
    token
        a unique instrument token
    initial_price
        initial_price for the ticker
    ticker_mode
        ticker mode; random or otherwise
    Note
    -----
    1) If ticker mode is random, price is generated based on
    random walk from normal distribution
    """

    name: str
    token: Optional[int] = None
    initial_price: float = 100
    mode: TickerMode = TickerMode.RANDOM
    orderbook: Optional[OrderBook]
    volume: Optional[int]
    _high: float = PrivateAttr()
    _low: float = PrivateAttr()
    _ltp: float = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._high = self.initial_price
        self._low = self.initial_price
        self._ltp = self.initial_price

    def _update_values(self, last_price: float):
        self._ltp = last_price
        self._high = max(self._high, last_price)
        self._low = min(self._low, last_price)

    @property
    def is_random(self) -> bool:
        """
        returns True if the mode is random else False
        """
        return True if self.mode == TickerMode.RANDOM else False

    @property
    def ltp(self) -> float:
        """
        Get the last price and update it
        """
        if self.is_random:
            diff = random.gauss(0, 1) * self._ltp * 0.01
            last_price = self._ltp + diff
            last_price = round(last_price * 20) / 20
            self._update_values(last_price)
        return self._ltp

    def update(self, last_price: float) -> float:
        """
        Update last price,high and low
        """
        self._update_values(last_price)
        return self._ltp

    def ohlc(self) -> OHLC:
        """
        Calculate the ohlc for this ticker
        """
        return OHLC(
            open=self.initial_price,
            high=self._high,
            low=self._low,
            close=self._ltp,
            last_price=self._ltp,
        )



class VQuote(OHLCV):
    orderbook: OrderBook


class VTrade(BaseModel):
    trade_id: str
    order_id: str
    symbol: str
    quantity: int
    price: float
    side: Side
    timestamp: Optional[pendulum.DateTime]

    class Config:
        validate_assignment = True

    @property
    def value(self) -> float:
        return self.side.value * self.quantity * self.price


class VOrder(BaseModel):
    order_id: str
    symbol: str
    quantity: float
    side: Side
    price: Optional[float]
    average_price: Optional[float]
    trigger_price: Optional[float]
    timestamp: Optional[pendulum.DateTime] = None
    exchange_order_id: Optional[str]
    exchange_timestamp: Optional[pendulum.DateTime]
    status_message: Optional[str]
    filled_quantity: float = 0
    pending_quantity: float = 0
    canceled_quantity: float = 0
    _delay: int = PrivateAttr()

    class Config:
        validate_assignment = True

    @validator('side', pre=True, always=True)
    def accept_buy_sell_as_side(cls, v):
        if isinstance(v, str):
            if v.lower()[0] == 'b':
                return Side.BUY
            elif v.lower()[0] == 's':
                return Side.SELL
            else:
                raise TypeError(f"{v} is not a valid side, should be buy or sell")
        else:
            return v

    def __init__(self, **data):
        super().__init__(**data)
        if self.timestamp is None:
            self.timestamp = pendulum.now(tz="local")
        q = utils.update_quantity(
            q=self.quantity,
            f=self.filled_quantity,
            p=self.pending_quantity,
            c=self.canceled_quantity,
        )
        self.filled_quantity = q.f
        self.pending_quantity = q.p
        self.canceled_quantity = q.c
        if self.average_price is None:
            self.average_price = 0
        self._delay = 1e6  # delay in microseconds

    def _modify_order_by_status(self, status: Status):
        """
        Modify an order quantity based on the given status
        """
        if status in (Status.CANCELED, Status.REJECTED):
            self.filled_quantity = 0
            self.pending_quantity = 0
            self.canceled_quantity = self.quantity
        elif status == Status.OPEN:
            self.filled_quantity = 0
            self.pending_quantity = self.pending_quantity
            self.canceled_quantity = 0
        elif status == Status.PARTIAL_FILL:
            a = random.randrange(1, int(self.quantity))
            b = self.quantity - a
            self.filled_quantity = a
            self.pending_quantity = 0
            self.canceled_quantity = b
        elif status == Status.PENDING:
            a = random.randrange(1, int(self.quantity))
            b = self.quantity - a
            self.filled_quantity = a
            self.pending_quantity = b
            self.canceled_quantity = 0
        else:
            self.filled_quantity = self.quantity
            self.pending_quantity = 0
            self.canceled_quantity = 0

    @property
    def is_past_delay(self) -> bool:
        """
        returns True is the order is past delay
        """
        if self.timestamp:
            expiry = self.timestamp.add(microseconds=self._delay)
            return True if pendulum.now(tz="local") > expiry else False
        else:
            return False

    @property
    def status(self) -> Status:
        if self.quantity == self.filled_quantity:
            return Status.COMPLETE
        elif self.quantity == self.canceled_quantity:
            if self.status_message:
                if str(self.status_message).upper().startswith("REJ"):
                    return Status.REJECTED
                else:
                    return Status.CANCELED
            else:
                return Status.CANCELED
        elif self.canceled_quantity > 0:
            if (self.canceled_quantity + self.filled_quantity) == self.quantity:
                return Status.PARTIAL_FILL
            else:
                return Status.PENDING
        elif self.pending_quantity > 0:
            if self.filled_quantity > 0:
                return Status.PENDING
            else:
                return Status.OPEN
        else:
            return Status.OPEN

    @property
    def value(self) -> float:
        """
        returns the value of the order
        negative means sell and positive means buy
        """
        if not self.average_price:
            if not self.price:
                average_price = 0.0
            else:
                average_price = self.price
        else:
            average_price = self.average_price
        return self.side.value * self.filled_quantity * average_price

    @property
    def is_done(self) -> bool:
        """
        whether the order is finished either by fully filled
        or canceled
        returns True if it is completed, False if pending
        """
        if self.quantity == self.filled_quantity:
            return True
        elif self.quantity == self.canceled_quantity:
            return True
        elif self.pending_quantity > 0:
            return False
        else:
            return True

    def modify_by_status(self, status: Status = Status.COMPLETE) -> bool:
        """
        Modify order by status
        returns True if the order is modified else False
        """
        if self.is_done:
            return False
        if self.is_past_delay:
            self._modify_order_by_status(status)
            return True
        else:
            return False


class VPosition(BaseModel):
    symbol: str
    buy_quantity: Optional[Union[int, float]]
    sell_quantity: Optional[Union[int, float]]
    buy_value: Optional[float]
    sell_value: Optional[float]

    class Config:
        validate_assignment = True

    @property
    def average_buy_price(self) -> float:
        """
        Get the average buy price
        returns 0 if there is no price or quantity
        """
        if self.buy_quantity and self.buy_value:
            return self.buy_value / self.buy_quantity
        else:
            return 0.0

    @property
    def average_sell_price(self) -> float:
        """
        Get the average sell price
        returns 0 if there is no price or quantity
        """
        if self.sell_quantity and self.sell_value:
            return self.sell_value / self.sell_quantity
        else:
            return 0.0

    @property
    def net_quantity(self) -> float:
        """
        Get the net quantity for the position
        negative indicates sell and positive indicates sell
        """
        buy_qty = self.buy_quantity if self.buy_quantity else 0
        sell_qty = self.sell_quantity if self.sell_quantity else 0
        return buy_qty - sell_qty

    @property
    def net_value(self) -> float:
        """
        Get the net value for the position
        negative indicates a net sell value and positive indicates a net buy value
        """
        buy_value = self.buy_value if self.buy_value else 0
        sell_value = self.sell_value if self.sell_value else 0
        return buy_value - sell_value


class VUser(BaseModel):
    userid: str
    name: Optional[str]
    orders: List[VOrder] = Field(default_factory=list)

    @validator("userid")
    def userid_should_be_upper(cls, v):
        return str(v).upper()

    def add(self, order: VOrder):
        """
        add an order to the user
        """
        self.orders.append(order)


class Response(BaseModel):
    status: ResponseStatus
    timestamp: Optional[pendulum.DateTime] = None

    class Config:
        validate_assignment = True

    def __init__(self, **data):
        super().__init__(**data)
        if self.timestamp is None:
            self.timestamp = pendulum.now(tz="local")


class OrderResponse(Response):
    error_msg: Optional[str] = None
    data: Optional[VOrder] = None

    class Config:
        validate_assignment = True


class AuthResponse(Response):
    user_id: str
    message: str = "Authentication successful"


class GenericResponse(OrderResponse):
    data: Optional[Any]


class LTPResponse(GenericResponse):
    data: Dict[str, Union[int, float]]


class OHLCVResponse(GenericResponse):
    data: Dict[str, OHLCV]


class QuoteResponse(GenericResponse):
    data: Dict[str, VQuote]


class OrderBookResponse(GenericResponse):
    data: Dict[str, OrderBook]


class PositionResponse(GenericResponse):
    data: List[VPosition]



from datetime import timezone
from typing import Optional, Dict, List, Type, Any, Union, Tuple, Callable
from omspy.base import Broker
from omspy.order import Order, CompoundOrder
from pydantic import PrivateAttr


class StopOrder(CompoundOrder):
    symbol: str
    side: str
    trigger_price: float
    price: float = 0.0
    quantity: int = 1
    disclosed_quantity: int = 0
    order_type: Optional[Tuple[str, str]] = None
    # TODO: Add order lock on modify

    def __init__(self, **data):
        super().__init__(**data)
        if self.order_type is None:
            self.order_type = ("LIMIT", "SL-M")
        side_map = {"buy": "sell", "sell": "buy"}
        base_order = Order(
            symbol=self.symbol,
            side=self.side,
            quantity=self.quantity,
            disclosed_quantity=self.disclosed_quantity,
            order_type=self.order_type[0],
            price=self.price,
            trigger_price=0,
        )

        cover_order = base_order.clone()
        cover_order.trigger_price = self.trigger_price
        cover_order.order_type = self.order_type[1]
        cover_order.side = side_map.get(cover_order.side)
        self.add(base_order)
        self.add(cover_order)


class StopLimitOrder(StopOrder):
    """
    Stop Loss Limit order
    stop_limit_price
        limit price for the stop order
    """

    stop_limit_price: float
    order_type: Tuple[str, str] = ("LIMIT", "SL")

    def __init__(self, order_type, **data):
        super().__init__(**data)
        self.orders[0].order_type = self.order_type[0]
        self.orders[-1].order_type = self.order_type[1]
        self.orders[-1].price = self.stop_limit_price


class TrailingStopOrder(StopOrder):
    """
    Trailing stop order
    trail_by
        trail_by in price
    """

    trail_by: float
    _next_trail: Optional[float] = PrivateAttr()
    _stop_loss: float = PrivateAttr()

    @property
    def sign(self) -> int:
        return 1 if self.side == "buy" else -1

    def __init__(self, **data):
        super().__init__(**data)
        self._stop_loss = self.trigger_price
        self._update_next_trail()

    def _update_next_trail(self):
        """
        Update trailing stop loss
        """
        price = self.orders[0].average_price if self.price == 0 else self.price
        if self.price > 0:
            self._next_trail = price + self.trail_by * self.sign * 1
        else:
            self._next_trail = 0

    @property
    def next_trail(self) -> float:
        return self._next_trail

    def run(self, ltp: float):
        """
        Update trailing stop
        """
        if self.next_trail == 0:
            self._update_next_trail()
        if self.next_trail > 0:
            if self.side == "buy":
                if ltp > self.next_trail:
                    # TODO: Trail to adjust to the nearest trail in case of jump in ltp
                    self._stop_loss += self.trail_by
                    self._next_trail += self.trail_by
                    self.orders[-1].modify(
                        broker=self.broker, trigger_price=self._stop_loss
                    )
            elif self.side == "sell":
                if ltp < self.next_trail:
                    self._stop_loss -= self.trail_by
                    self._next_trail -= self.next_trail
                    self.orders[-1].modify(
                        broker=self.broker, trigger_price=self._stop_loss
                    )


class TargetOrder(StopOrder):
    """
    Exit an order when the target price is hit
    target
        target price to exit order
    Note
    -----
    1) The existing stop loss order is converted into a MARKET order when the target price is hit
    """

    target: float
    order_type: Tuple[str, str] = ("LIMIT", "SL-M")

    def __init__(self, **data):
        super().__init__(**data)

    def run(self, ltp: float):
        """
        Update and exit if target is hit
        """
        price = self.price if self.price > 0 else self.orders[0].average_price
        if price > 0:
            if self.side == "buy":
                if ltp >= self.target:
                    self.orders[-1].modify(broker=self.broker, order_type="MARKET")
            elif self.side == "sell":
                if ltp <= self.target:
                    self.orders[-1].modify(broker=self.broker, order_type="MARKET")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['omspy', 'omspy.algos', 'omspy.brokers', 'omspy.orders', 'omspy.simulation']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'sqlite-utils>=3.22.1,<4.0.0']

setup_kwargs = {
    'name': 'omspy',
    'version': '0.15.0',
    'description': '',
    'long_description': '# Omspy\n\nomspy is a broker agnostic order management system with a common api, advanced order types and more\n',
    'author': 'Ubermensch',
    'author_email': 'uberdeveloper001@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/uberdeveloper/omspy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from pydantic import BaseModel, validator, Field, PrivateAttr, Json
from datetime import timezone
from typing import (
    Optional,
    Dict,
    List,
    Type,
    Any,
    Union,
    Tuple,
    Callable,
    Set,
    Hashable,
)
import uuid
import pendulum
import sqlite3
import logging
from collections import Counter, defaultdict
from collections.abc import Iterable
from omspy.base import *
from copy import deepcopy
from sqlite_utils import Database
from omspy.models import OrderLock


def get_option(spot: float, num: int = 0, step: float = 100.0) -> float:
    """
    Get the option price given number of strikes
    spot
        spot price of the instrument
    num
        number of strikes farther
    step
        step size of the option
    Note
    ----
    1. By default, the ATM option is fetched
    """
    v = round(spot / step)
    return v * (step + num)


def create_db(dbname: str = ":memory:") -> Union[Database, None]:
    """
    Create a sqlite3 database for the orders and return the connection
    dbname
        name of the database
        default in-memory database
    """
    try:
        con = sqlite3.connect(dbname)
        with con:
            con.execute(
                """create table orders
                           (
                           symbol text, side text, quantity integer,
                           id text primary key, parent_id text, timestamp text,
                           order_type text, broker_timestamp text,
                           exchange_timestamp text, order_id text,
                           exchange_order_id text, price real,
                           trigger_price real, average_price real,
                           pending_quantity integer, filled_quantity integer,
                           cancelled_quantity integer, disclosed_quantity integer,
                           validity text, status text,
                           expires_in integer, timezone text,
                           client_id text, convert_to_market_after_expiry text,
                           cancel_after_expiry text, retries integer, max_modifications integer,
                           exchange text, tag string, can_peg integer,
                           pseudo_id string, strategy_id string, portfolio_id string,
                           JSON text, error text, is_multi integer,
                           last_updated_at text
                           )"""
            )
            return Database(con)
    except Exception as e:
        logging.error(e)
        return None


class Order(BaseModel):
    """
    The basic Order Class
    _attrs
        attributes to update when data received from broker
    _frozen_attrs
        attributes frozen; cannot be changed when modifying orders
    """

    symbol: str
    side: str
    quantity: int = 1
    id: Optional[str] = None
    parent_id: Optional[str] = None
    timestamp: Optional[pendulum.DateTime] = None
    order_type: str = "MARKET"
    broker_timestamp: Optional[pendulum.DateTime] = None
    exchange_timestamp: Optional[pendulum.DateTime] = None
    order_id: Optional[str] = None
    exchange_order_id: Optional[str] = None
    price: Optional[float] = None
    trigger_price: float = 0.0
    average_price: float = 0.0
    pending_quantity: Optional[int] = None
    filled_quantity: int = 0
    cancelled_quantity: int = 0
    disclosed_quantity: int = 0
    validity: str = "DAY"
    status: Optional[str] = None
    expires_in: int = 0
    timezone: str = "local"
    client_id: Optional[str] = None
    convert_to_market_after_expiry: bool = False
    cancel_after_expiry: bool = True
    retries: int = 0
    max_modifications: int = 10
    exchange: Optional[str] = None
    tag: Optional[str] = None
    connection: Optional[Database] = None
    can_peg: bool = True
    pseudo_id: Optional[str] = None
    strategy_id: Optional[str] = None
    portfolio_id: Optional[str] = None
    JSON: Optional[Json] = None
    error: Optional[str] = None
    is_multi: bool = False
    last_updated_at: Optional[pendulum.DateTime] = None
    _num_modifications: int = 0
    _attrs: Tuple[str, ...] = (
        "exchange_timestamp",
        "exchange_order_id",
        "status",
        "filled_quantity",
        "pending_quantity",
        "disclosed_quantity",
        "average_price",
    )
    _exclude_fields: Set[str] = {"connection"}
    _lock: Optional[OrderLock] = None
    _frozen_attrs: Set[str] = {"symbol", "side"}

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True

    def __init__(self, **data) -> None:
        super().__init__(**data)
        from omspy.base import Broker

        if not (self.id):
            self.id = uuid.uuid4().hex
        tz = self.timezone
        if not (self.timestamp):
            self.timestamp = pendulum.now(tz=tz)
        self.pending_quantity = self.quantity
        if self.expires_in == 0:
            self.expires_in = (
                pendulum.today(tz=tz).end_of("day") - pendulum.now(tz=tz)
            ).seconds
        else:
            self.expires_in = abs(self.expires_in)
        if self._lock is None:
            self._lock = OrderLock()

    @validator("quantity", always=True, allow_reuse=True)
    def quantity_not_negative(cls, v):
        if v < 0:
            raise ValueError("quantity must be positive")
        return v

    @property
    def is_complete(self) -> bool:
        if self.quantity == self.filled_quantity:
            return True
        elif self.status == "COMPLETE":
            return True
        elif (self.filled_quantity + self.cancelled_quantity) == self.quantity:
            return True
        else:
            return False

    @property
    def is_pending(self) -> bool:
        quantity = self.filled_quantity + self.cancelled_quantity
        # Order not pending if it is complete/canceled or rejected
        # irrespective of the filled and remaining quantity
        if self.status in ("COMPLETE", "CANCELED", "REJECTED"):
            return False
        elif quantity < self.quantity:
            return True
        else:
            return False

    @property
    def is_done(self) -> bool:
        """
        returns True if the order is either COMPLETE or CANCELED or REJECTED else False
        """
        if self.is_complete:
            return True
        elif self.status in ("CANCELLED", "CANCELED", "REJECTED"):
            return True
        else:
            return False

    @property
    def time_to_expiry(self) -> int:
        now = pendulum.now(tz=self.timezone)
        ts = self.timestamp
        return max(0, self.expires_in - (now - ts).seconds)

    @property
    def time_after_expiry(self) -> int:
        now = pendulum.now(tz=self.timezone)
        ts = self.timestamp
        return max(0, (now - ts).seconds - self.expires_in)

    @property
    def has_expired(self) -> bool:
        return True if self.time_to_expiry == 0 else False

    @property
    def has_parent(self) -> bool:
        return True if self.parent_id else False

    @property
    def lock(self) -> OrderLock:
        if self._lock is None:
            self._lock = OrderLock()
        return self._lock

    def _get_other_args_from_attribs(
        self, broker: Any, attribute: str, attribs_to_copy: Optional[Iterable] = None
    ) -> Dict[str, str]:
        """
        Get other arguments for the order from attributes
        broker
            valid broker instance
        attribute
            attribute to search for in broker
        attribs_to_copy
            extra attributes to be copied
        Note
        ----
        1) The broker instance is first searched for the valid attribute and it is overriden with attribs_to_copy
        """
        if attribs_to_copy is None:
            attribs_to_copy = set()
        else:
            # Convert any iterable
            attribs_to_copy = set([x for x in attribs_to_copy])
        if hasattr(broker, attribute):
            attribs = getattr(broker, attribute)
            for attrib in attribs:
                attribs_to_copy.add(attrib)
        other_args = dict()
        if attribs_to_copy:
            for key in attribs_to_copy:
                if hasattr(self, key):
                    value = getattr(self, key)
                    if value:
                        other_args[key] = value
        return other_args

    def update(self, data: Dict[str, Any], save: bool = True) -> bool:
        """
        Update order based on information received from broker
        data
            data to update as dictionary
        returns True if update is done
        Note
        ----
        1) Information is updated only for those keys specified in attrs
        2) Information is updated only when the order is still pending; completed/rejected/canceled orders not updated
        3) Update pending quantity if it is not in data
        """
        if not (self.is_done):
            for att in self._attrs:
                val = data.get(att)
                if val:
                    setattr(self, att, val)
            self.last_updated_at = pendulum.now(tz=self.timezone)
            if not ("pending_quantity" in data):
                self.pending_quantity = self.quantity - self.filled_quantity
            if self.connection and save:
                self.save_to_db()
            return True
        else:
            return False

    def execute(
        self, broker: Any, attribs_to_copy: Optional[Set] = None, **kwargs
    ) -> Optional[str]:
        """
        Execute an order on a broker, place a new order
        kwargs
            Additional arguments to the order
        Note
        ----
        Only new arguments added to the order in keyword arguments
        """
        # Do not place a new order if this order is complete or has order_id

        from omspy.base import Broker as base_broker

        other_args = self._get_other_args_from_attribs(
            broker, attribute="attribs_to_copy_execute", attribs_to_copy=attribs_to_copy
        )
        if not (self.is_complete) and not (self.order_id):
            order_args = {
                "symbol": self.symbol.upper(),
                "side": self.side.upper(),
                "order_type": self.order_type.upper(),
                "quantity": self.quantity,
                "price": self.price,
                "trigger_price": self.trigger_price,
                "disclosed_quantity": self.disclosed_quantity,
            }
            dct = {k: v for k, v in kwargs.items() if k not in order_args.keys()}
            order_args.update(other_args)
            order_args.update(dct)
            order_id = broker.order_place(**order_args)
            self.order_id = order_id
            if self.connection:
                self.save_to_db()
            return order_id
        else:
            return self.order_id

    def modify(
        self, broker: Any, attribs_to_copy: Optional[Tuple] = None, **kwargs
    ) -> None:
        """
        Modify an existing order
        broker
            broker to execute the order
        attribs
            attributes to be copied from the order
        kwargs
            keyword arguments for modification
        Note
        ----
        1)resolution for order args - default arguments are created for modify, attribs_to_copy are added next and finally kwargs are updated. If the same attribute is found in all three, the user provided kwargs wins
        """
        if not (self.lock.can_modify):
            logging.debug(
                f"Order not modified since lock is modified till {self.lock.modification_lock_till}"
            )
            return
        other_args = self._get_other_args_from_attribs(
            broker, attribute="attribs_to_copy_modify", attribs_to_copy=attribs_to_copy
        )
        args_to_add = dict()
        keys = [
            "order_id",
            "quantity",
            "price",
            "trigger_price",
            "order_type",
            "disclosed_quantity",
        ]
        for k, v in kwargs.items():
            if k not in self._frozen_attrs:
                if hasattr(self, k):
                    setattr(self, k, v)
                    if k not in keys:
                        args_to_add[k] = v
                else:
                    other_args[k] = v
        order_args = {
            "order_id": self.order_id,
            "quantity": self.quantity,
            "price": self.price,
            "trigger_price": self.trigger_price,
            "order_type": self.order_type.upper(),
            "disclosed_quantity": self.disclosed_quantity,
        }
        order_args.update(other_args)
        order_args.update(args_to_add)
        if self._num_modifications < self.max_modifications:
            broker.order_modify(**order_args)
            self._num_modifications += 1
        else:
            logging.info(f"Maximum number of modifications exceeded")

    def cancel(self, broker: Any, attribs_to_copy: Optional[Set] = None) -> None:
        """
        Cancel an existing order
        """
        if not (self.lock.can_cancel):
            logging.debug(
                f"Order not canceled since lock is modified till {self.lock.cancellation_lock_till}"
            )
            return
        other_args = self._get_other_args_from_attribs(
            broker, attribute="attribs_to_copy_cancel", attribs_to_copy=attribs_to_copy
        )
        if self.order_id is not None:
            broker.order_cancel(order_id=self.order_id, **other_args)

    def save_to_db(self) -> bool:
        """
        save or update the order to db
        """
        if self.connection:
            values = self.dict(exclude=self._exclude_fields)
            self.connection["orders"].upsert(values, pk="id")
            return True
        else:
            logging.info("No valid database connection")
            return False

    def clone(self):
        """
        Clone the order with a new order id
        Note
        ----
        returns a copy of the new order with a new
        order_id. parent_id is not copied
        """
        dct = self.dict(exclude={"id", "parent_id", "timestamp"})
        order = Order(**dct)
        return order

    def add_lock(self, code: int, seconds: float):
        """
        Create a lock on modify or cancel function
        code
            1 to lock modify and 2 to lock cancel
        seconds
            duration to lock the function
        Note
        ----
        1) Lock is created only on modify or cancel
        """
        if code == 1:
            self.lock.modify(seconds=seconds)
        elif code == 2:
            self.lock.cancel(seconds=seconds)


class CompoundOrder(BaseModel):
    """
    A collection of orders
    Note
    ----
    1) Indexes are added automatically based on the highest key value
    2) An error is raised if the index value is already used
    """

    broker: Any
    id: Optional[str] = None
    ltp: defaultdict = Field(default_factory=defaultdict)
    orders: List[Order] = Field(default_factory=list)
    connection: Optional[Database] = None
    order_args: Optional[Dict] = None
    _index: Dict[int, Order] = PrivateAttr(default_factory=defaultdict)
    _keys: Dict[Hashable, Order] = PrivateAttr(default_factory=defaultdict)

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True

    def __init__(self, **data) -> None:
        super().__init__(**data)
        if not (self.id):
            self.id = uuid.uuid4().hex
        if self.order_args is None:
            self.order_args = {}
        if self.orders:
            for i, o in enumerate(self.orders):
                self._index[i] = o

    @property
    def count(self) -> int:
        """
        return the number of orders
        """
        return len(self.orders)

    @property
    def positions(self) -> Counter:
        """
        return the positions as a dictionary
        """
        c: Counter = Counter()
        for order in self.orders:
            symbol = order.symbol
            qty = order.filled_quantity
            side = str(order.side).lower()
            sign = -1 if side == "sell" else 1
            qty = qty * sign
            c.update({symbol: qty})
        return c

    def _get_next_index(self) -> int:
        idx = max(self._index.keys()) + 1 if self._index else 0
        return idx

    def _get_by_key(self, key: Hashable) -> Union[Order, None]:
        return self._keys.get(key)

    def _get_by_index(self, index: int) -> Union[Order, None]:
        return self._index.get(index)

    def get(self, key: Hashable) -> Union[Order, None]:
        """
        Get the order by key or index
        key
            index or key
        returns Order if available or None
        Note
        ----
        1) key is first searched in key and then searched in index
        2) index starts at zero
        """
        order = self._get_by_key(key)
        if order is None:
            try:
                if isinstance(key, (str, int)):
                    int_key = int(key)
                    return self._get_by_index(int_key)
                else:
                    return None
            except Exception as e:
                return None
        else:
            return order

    def add_order(self, **kwargs) -> Optional[str]:
        kwargs["parent_id"] = self.id
        index = kwargs.pop("index", self._get_next_index())
        key = kwargs.pop("key", None)
        if not (kwargs.get("connection")):
            kwargs["connection"] = self.connection
        if index in self._index:
            raise IndexError("Order already assigned to this index")
        if key:
            if key in self._keys:
                raise KeyError("Order already assigned to this key")
        order = Order(**kwargs)
        self.orders.append(order)
        self._index[index] = order
        if key:
            self._keys[key] = order
        order.save_to_db()
        return order.id

    def _average_price(self, side: str = "buy") -> Dict[str, float]:
        """
        Get the average price for all the instruments
        side
            side to calculate average price - buy or sel
        """
        side = str(side).lower()
        value_counter: Counter = Counter()
        quantity_counter: Counter = Counter()
        for order in self.orders:
            order_side = str(order.side).lower()
            if side == order_side:
                symbol = order.symbol
                price = order.average_price
                quantity = order.filled_quantity
                value = price * quantity
                value_counter.update({symbol: value})
                quantity_counter.update({symbol: quantity})
        dct: defaultdict = defaultdict()
        for v in value_counter:
            numerator = value_counter.get(v)
            denominator = quantity_counter.get(v)
            if numerator and denominator:
                dct[v] = numerator / denominator
        return dct

    @property
    def average_buy_price(self) -> Dict[str, float]:
        return self._average_price(side="buy")

    @property
    def average_sell_price(self) -> Dict[str, float]:
        return self._average_price(side="sell")

    def update_orders(self, data: Dict[str, Dict[str, Any]]) -> Dict[str, bool]:
        """
        Update all orders
        data
            data as dictionary with key as broker order_id
        returns a dictionary with order_id and update status as boolean
        """
        dct: Dict[str, bool] = {}
        for order in self.pending_orders:
            order_id = str(order.order_id)
            status = order.status
            if order_id in data:
                d = data.get(order_id)
                if d:
                    order.update(d)
                    dct[order_id] = True
                else:
                    dct[order_id] = False
            else:
                dct[order_id] = False
        return dct

    def _total_quantity(self) -> Dict[str, Counter]:
        """
        Get the total buy and sell quantity by symbol
        """
        buy_counter: Counter = Counter()
        sell_counter: Counter = Counter()
        for order in self.orders:
            side = order.side.lower()
            symbol = order.symbol
            quantity = abs(order.filled_quantity)
            if side == "buy":
                buy_counter.update({symbol: quantity})
            elif side == "sell":
                sell_counter.update({symbol: quantity})
        return {"buy": buy_counter, "sell": sell_counter}

    @property
    def buy_quantity(self) -> Counter:
        return self._total_quantity()["buy"]

    @property
    def sell_quantity(self) -> Counter:
        return self._total_quantity()["sell"]

    def update_ltp(self, last_price: Dict[str, float]):
        """
        Update ltp for the given symbols
        last_price
            dictionary with symbol as key and last price as value
        returns the ltp for all the symbols
        Note
        ----
        1. Last price is updated for all given symbols irrespective of
        orders placed
        """
        for symbol, ltp in last_price.items():
            self.ltp[symbol] = ltp
        return self.ltp

    @property
    def net_value(self) -> Counter:
        """
        Return the net value by symbol
        """
        c: Counter = Counter()
        for order in self.orders:
            symbol = order.symbol
            side = str(order.side).lower()
            sign = -1 if side == "sell" else 1
            value = order.filled_quantity * order.average_price * sign
            c.update({symbol: value})
        return c

    @property
    def mtm(self) -> Counter:
        c: Counter = Counter()
        net_value = self.net_value
        positions = self.positions
        ltp = self.ltp
        for symbol, value in net_value.items():
            c.update({symbol: -value})
        for symbol, quantity in positions.items():
            v = quantity * ltp.get(symbol, 0)
            c.update({symbol: v})
        return c

    @property
    def total_mtm(self) -> float:
        return sum(self.mtm.values())

    def execute_all(self, **kwargs):
        for order in self.orders:
            order_args = deepcopy(self.order_args)
            order_args.update(kwargs)
            order.execute(broker=self.broker, **order_args)

    def check_flags(self) -> None:
        """
        Check for flags on each order and take suitable action
        """
        for order in self.orders:
            if (order.is_pending) and (order.has_expired):
                if order.convert_to_market_after_expiry:
                    order.order_type = "MARKET"
                    order.modify(self.broker)
                elif order.cancel_after_expiry:
                    order.cancel(broker=self.broker)

    @property
    def completed_orders(self) -> List[Order]:
        return [order for order in self.orders if order.is_complete]

    @property
    def pending_orders(self) -> List[Order]:
        return [order for order in self.orders if order.is_pending]

    def add(
        self, order: Order, index: Optional[int] = None, key: Optional[Hashable] = None
    ) -> Optional[str]:
        """
        Add an order to the existing compound order
        """
        order.parent_id = self.id
        if not (order.connection):
            order.connection = self.connection
        if not (order.id):
            order.id = uuid.uuid4().hex
        if index is None:
            index = self._get_next_index()
        index = int(index)
        if index in self._index:
            raise IndexError("Order already assigned to this index")
        if key:
            if key in self._keys:
                raise KeyError("Order already assigned to this key")
        self.orders.append(order)
        self._index[index] = order
        if key:
            self._keys[key] = order
        order.save_to_db()
        return order.id

    def save(self) -> None:
        """
        Save all orders to database
        """
        if self.count > 0:
            for order in self.orders:
                order.save_to_db()


class OrderStrategy(BaseModel):
    """
    An order strategy is a list of strategies that
    is made up of a list of compound orders
    """

    broker: Any
    id: Optional[str] = None
    ltp: defaultdict = Field(default_factory=defaultdict)
    orders: List[CompoundOrder] = Field(default_factory=list)

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True

    def __init__(self, **data) -> None:
        super().__init__(**data)
        if not (self.id):
            self.id = uuid.uuid4().hex

    @property
    def positions(self) -> Counter:
        c: Counter = Counter()
        for order in self.orders:
            pos = order.positions
            c.update(pos)
        return c

    def update_ltp(self, last_price: Dict[str, float]):
        for symbol, ltp in last_price.items():
            self.ltp[symbol] = ltp
        for order in self.orders:
            order.update_ltp(last_price)
        return self.ltp

    def update_orders(self, data: Dict[str, Dict[str, Any]]) -> None:
        """
        Update all orders
        data
            data as dictionary with key as broker order_id
        """
        for order in self.orders:
            order.update_orders(data)

    @property
    def mtm(self) -> Counter:
        c: Counter = Counter()
        for order in self.orders:
            mtm = order.mtm
            c.update(mtm)
        return c

    def run(self, ltp: Dict[str, float]) -> None:
        """
        Run all orders with the given data
        ltp
            last price data as a dictionary
        """
        for order in self.orders:
            if hasattr(order, "run"):
                if callable(getattr(order, "run")):
                    order.run(ltp)

    def add(self, order: CompoundOrder) -> None:
        """
        Add a compound order to the existing strategy
        """
        self.orders.append(order)

    def save(self) -> None:
        """
        Save all orders to database
        Note
        ----
        1) Orders are saved based on the preferences of each compound order; so this doesn't save everything
        """
        for order in self.orders:
            order.save()
        pass



import random
import uuid
import logging
from functools import wraps
from typing import Optional, Dict, Set, List, Union, Any, Callable
from omspy.models import OrderBook, Quote
from pydantic import BaseModel, PrivateAttr, confloat, ValidationError, Field
from enum import Enum
from collections import defaultdict
from collections.abc import Iterable
from omspy.simulation.models import (
    OrderResponse,
    ResponseStatus,
    VOrder,
    VTrade,
    OHLCV,
    Side,
    Status,
    VQuote,
    VPosition,
    VUser,
    TickerMode,
    Ticker,
)

SUCCESS = ResponseStatus.SUCCESS
FAILURE = ResponseStatus.FAILURE


def user_response(f: Callable):
    """
    override a function with the response provided by the user
    if the user includes the `response` keyword,
    user response would be returns instead of the function response
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        if "response" in kwargs:
            return kwargs.pop("response")
        else:
            return f(*args, **kwargs)

    return wrapper


def _iterate_method(
    method: Callable, symbol: Union[str, Iterable], **kwargs
) -> Dict[str, Any]:
    """
    iterate the given method if the symbol is an iterable else return the value
    """
    if isinstance(symbol, str):
        return method(symbol, **kwargs)
    elif isinstance(symbol, Iterable):
        dct = dict()
        for s in symbol:
            val = method(s, **kwargs)
            if val:
                dct.update(val)
        return dct
    else:
        return dict()


def generate_price(start: int = 100, end: int = 110) -> int:
    """
    Generate a random price in the given range between start and end
    start
        starting value
    end
        ending value
    Note
    ----
    1) If the start value is greater than end value, the values are swapped
    """
    if start > end:
        start, end = end, start
    return random.randrange(start, end)


def generate_orderbook(
    bid: float = 100.0,
    ask: float = 100.05,
    depth: int = 5,
    tick: float = 0.01,
    quantity: int = 100,
) -> OrderBook:
    """
    generate a fake orderbook
    bid
        bid price
    ask
        ask price
    depth
        depth of the orderbook
    tick
        difference in price between orders
    quantity
        average quantity of orders per price quote

    Note
    ----
    1) orderbook is generated with a uniform tick difference between subsequent quotes
    2) quantity is averaged between value quantity/2 and quantity * 1.5
    using randrange function
    3) num of orders is randomly picked between 5 to 15
    4) if bid price is greater than ask, the values are swapped
    """
    if bid > ask:
        bid, ask = ask, bid
    asks = []
    bids = []
    q1, q2 = int(quantity * 0.5), int(quantity * 1.5)
    for i in range(depth):
        bid_qty = random.randrange(q1, q2)
        ask_qty = random.randrange(q1, q2)
        b = Quote(
            price=bid - i * tick,
            quantity=bid_qty,
            orders_count=min(random.randrange(5, 15), bid_qty),
        )
        a = Quote(
            price=ask + i * tick,
            quantity=ask_qty,
            orders_count=min(random.randrange(5, 15), ask_qty),
        )
        bids.append(b)
        asks.append(a)
    return OrderBook(ask=asks, bid=bids)


def generate_ohlc(start: int = 100, end: int = 110, volume: int = 10000) -> OHLCV:
    """
    Generate random open, high, low, close prices
    start
        start value for price generation
    end
        end value for price generation
    volume
        value for volume
    returns open, high, low, close, last price
    and volume by default
    Note
    ----
    1) ohlc is generated between start and end values
    2) volume is generated based on given value
    """
    if start > end:
        start, end = end, start
    a = random.randrange(start, end)
    b = random.randrange(start, end)
    high, low = max(a, b), min(a, b)
    o = random.randrange(low, high)
    c = random.randrange(low, high)
    ltp = random.randrange(low, high)
    if volume > 0:
        v = random.randrange(int(volume * 0.5), int(volume * 2))
    else:
        v = random.randrange(1000, 200000)
    return OHLCV(open=o, high=high, low=low, close=c, volume=v, last_price=ltp)


class FakeBroker(BaseModel):
    """
    A fake instance to generate random stock data
    """

    name: str = "faker"
    _symbols: List[str] = [
        "AXP",
        "AAPL",
        "CSCO",
        "PG",
        "V",
        "MMM",
        "JPM",
        "HD",
        "CVX",
        "GS",
        "DOW",
    ]

    def _create_order_args(self, **kwargs) -> Dict[str, Any]:
        """
        Create order arguments from the list of
        keyword arguments
        """
        if "symbol" not in kwargs:
            kwargs["symbol"] = random.choice(self._symbols)
        if "quantity" not in kwargs:
            kwargs["quantity"] = random.randrange(10, 10000)
        if "price" not in kwargs:
            kwargs["price"] = random.randrange(1, 1000)
        if "side" not in kwargs:
            kwargs["side"] = random.choice([Side.BUY, Side.SELL])
        return kwargs

    def _get_random_symbols(self, n:Optional[int]=None)->List[str]:
        """
        get random symbols
        """
        if n is None:
            n = random.randrange(1, len(self._symbols))
        symbols = random.choices(self._symbols, k=n)
        return symbols

    def _ltp(self, symbol: str, **kwargs) -> Dict[str, Union[float, int]]:
        """
        get some random last traded price for the instrument
        symbol
            symbol name
        kwargs
            can provide start and end arguments to generate price within the range
        """
        price = generate_price(**kwargs)
        return {symbol: price}

    @user_response
    def ltp(
        self, symbol: Union[str, Iterable], **kwargs
    ) -> Dict[str, Union[float, int]]:
        """
        get some random last traded price for the given symbols
        symbol
            symbol could be a single symbol or a list of tuple of symbols
        kwargs
            can provide start and end arguments to generate price within the range
        """
        return _iterate_method(self._ltp, symbol, **kwargs)

    def _orderbook(self, symbol: str, **kwargs) -> Dict[str, OrderBook]:
        """
        generate a random orderbook
        """
        orderbook = generate_orderbook(**kwargs)
        return {symbol: orderbook}

    @user_response
    def orderbook(self, symbol: Union[str, Iterable], **kwargs) -> Dict[str, OrderBook]:
        """
        generate a random orderbook
        symbol
            symbol or list of symbols
        kwargs
            keyword arguments for the generate_orderbook funtion
        """
        return _iterate_method(self._orderbook, symbol, **kwargs)

    def _ohlc(self, symbol: str, **kwargs) -> Dict[str, OHLCV]:
        """
        generate ohlc prices
        """
        values = generate_ohlc(**kwargs)
        return {symbol: values}

    @user_response
    def ohlc(self, symbol: Union[str, Iterable], **kwargs) -> Dict[str, OHLCV]:
        """
        generate ohlc prices
        symbol
            symbol or list of symbols
        kwargs
            keyword arguments for the generate_ohlc funtion
        """
        return _iterate_method(self._ohlc, symbol, **kwargs)

    def _quote(self, symbol: str, **kwargs) -> Dict[str, VQuote]:
        """
        generate a detailed quote with ohlcv and orderbook
        start
            start price of the symbol
        end
            end price of the symbol
        volume
            volume for ohlc
        depth
            depth of the orderbook
        tick
            difference in price between orders
        quantity
            average quantity of orders per price quote
        Note
        -----
        1) ask and bid price are derived from start and end prices
        """
        start = kwargs.get("start", 100)
        end = kwargs.get("end", 110)
        volume = kwargs.get("volume", 1e4)
        depth = kwargs.get("depth", 5)
        tick = kwargs.get("tick", 0.01)
        quantity = kwargs.get("quantity", 100)
        ohlc = generate_ohlc(start=start, end=end, volume=volume)
        bid = generate_price(start=int(ohlc.low), end=int(ohlc.high))
        ask = bid + tick
        orderbook = generate_orderbook(
            ask=ask, bid=bid, depth=depth, tick=tick, quantity=quantity
        )
        quote = VQuote(orderbook=orderbook, **ohlc.dict())
        return {symbol: quote}

    @user_response
    def quote(self, symbol: Union[str, Iterable], **kwargs) -> Dict[str, VQuote]:
        """
        generate a detailed quote with ohlcv and orderbook
        symbol
            symbol or list of symbols
        list of kwargs
        start
            start price of the symbol
        end
            end price of the symbol
        volume
            volume for ohlc
        depth
            depth of the orderbook
        tick
            difference in price between orders
        quantity
            average quantity of orders per price quote
        Note
        -----
        1) ask and bid price are derived from start and end prices
        """
        return _iterate_method(self._quote, symbol, **kwargs)

    @user_response
    def order_place(self, **kwargs) -> VOrder:
        """
        Place an order with the broker
        """
        status: Optional[Status] = kwargs.get("s")
        order_args = self._create_order_args(**kwargs)
        quantity = order_args["quantity"]
        order_args["filled_quantity"] = quantity
        if status:
            if status in (Status.CANCELED, Status.REJECTED):
                order_args.update(dict(filled_quantity=0, canceled_quantity=quantity))
            elif status == Status.OPEN:
                order_args.update(dict(filled_quantity=0, pending_quantity=quantity))
            elif status == Status.PARTIAL_FILL:
                a = random.randrange(1, quantity)
                b = quantity - a
                order_args.update(dict(filled_quantity=a, canceled_quantity=b))
            elif status == Status.PENDING:
                a = random.randrange(1, quantity)
                b = quantity - a
                order_args.update(dict(filled_quantity=a, pending_quantity=b))
        order_id = uuid.uuid4().hex
        order_args.update(kwargs)
        return VOrder(order_id=order_id, **order_args)

    @user_response
    def order_modify(self, **kwargs) -> VOrder:
        """
        Modify an order with the broker
        All orders are returned with status OPEN
        """
        modify_args = self._create_order_args(**kwargs)
        quantity = modify_args["quantity"]
        order_id = modify_args.pop("order_id", uuid.uuid4().hex)
        modify_args["pending_quantity"] = quantity
        return VOrder(order_id=order_id, **modify_args)

    @user_response
    def order_cancel(self, **kwargs) -> VOrder:
        """
        Cancel an order with the broker
        All orders are returned with status CANCELED with
        entire quantity of the orders being CANCELED
        """
        cancel_args = self._create_order_args(**kwargs)
        quantity = cancel_args["quantity"]
        order_id = cancel_args.pop("order_id", uuid.uuid4().hex)
        cancel_args["canceled_quantity"] = quantity
        return VOrder(order_id=order_id, **cancel_args)

    @user_response
    def positions(self, symbols: Optional[List[str]] = None) -> List[VPosition]:
        """
        Generate some fake positions
        symbols
            symbols for which positions are to be generated
        """
        if not symbols:
            n = random.randrange(1, len(self._symbols))
            symbols = random.choices(self._symbols, k=n)
        symbols = list(set(symbols))  # To remove duplicates
        positions = []
        for symbol in symbols:
            bq = random.randrange(0, 1000)
            sq = random.randrange(0, 1000)
            bv = random.randrange(10, 3000)
            sv = random.randrange(int(bv * 0.5), int(bv * 2))
            position = VPosition(
                symbol=symbol,
                buy_quantity=bq,
                sell_quantity=sq,
                buy_value=bv,
                sell_value=sv,
            )
            positions.append(position)
        return positions

    @user_response
    def orders(self, symbols: Optional[List[str]] = None) -> List[VOrder]:
        """
        Generate some fake orders
        symbols
            symbol for which fake orders are to be generated
        """
        if not symbols:
            symbols = self._get_random_symbols()
        orders = []
        for symbol in symbols:
            order_id = uuid.uuid4().hex
            quantity = random.randrange(10, 100)
            price = round(random.random() * random.randrange(10, 100), 2)
            order = VOrder(
                order_id=order_id,
                symbol=symbol,
                quantity=quantity,
                filled_quantity=quantity,
                side=random.choice(list(Side)),
                price=price,
                average_price=price,
            )
            orders.append(order)
        return orders

    @user_response
    def trades(self, symbols: Optional[List[str]] = None) -> List[VTrade]:
        """
        Generate some fake trades
        symbols
            symbol for which fake trades are to be generated
        """
        if not symbols:
            n = random.randrange(1, len(self._symbols)) * 2
            symbols = self._get_random_symbols(n)
        trades = []
        for symbol in symbols:
            order_id = uuid.uuid4().hex
            trade_id = uuid.uuid4().hex
            quantity = random.randrange(10, 100)
            price = round(random.random() * random.randrange(10, 100), 2)
            trade = VTrade(
                trade_id=trade_id,
                order_id=order_id,
                symbol=symbol,
                quantity=quantity,
                side=random.choice(list(Side)),
                price=price,
            )
            trades.append(trade)
        return trades


class VirtualBroker(BaseModel):
    """
    A virtual broker instance mimicking a real broker
    name
        name to identify this broker
    tickers
        list of stock tickers
    users
        list of users
    failure_rate
        the failure rate for responses
    """

    name: str = "VBroker"
    tickers: Dict[str, Ticker] = Field(default_factory=dict)
    users: List[VUser] = Field(default_factory=list)
    failure_rate: float = Field(ge=0, le=1, default=0.001)
    _orders: Dict[str, VOrder] = PrivateAttr()
    _clients: Set[str] = PrivateAttr()
    _delay: int = PrivateAttr()  # delay in microseconds for updating orders

    class Config:
        validate_assignment = True

    def __init__(self, **data):
        super().__init__(**data)
        self._orders = defaultdict(list)
        self._clients = set()
        self._delay = 1e6

    @property
    def clients(self) -> Set[str]:
        return self._clients

    @property
    def is_failure(self) -> bool:
        """
        return whether the response should be a success or failure
        Note
        ----
        1) status is determined based on the failure rate
        """
        num = random.random()
        if num < self.failure_rate:
            return True
        else:
            return False

    def get(
        self, order_id: str, status: Status = Status.COMPLETE
    ) -> Union[VOrder, None]:
        """
        get the order
        """
        order: VOrder = self._orders.get(order_id)
        if order:
            order.modify_by_status(status)
            return order
        else:
            return None

    def add_user(self, user: VUser) -> bool:
        """
        add a new user
        returns True if an user is added successfully else False
        """
        if user.userid in self.clients:
            logging.warning(f"User {user.userid} already exists")
            return False
        else:
            self._clients.add(user.userid)
            self.users.append(user)
            return True

    def order_place(self, **kwargs) -> Union[OrderResponse, Dict[Any, Any]]:
        if "response" in kwargs:
            return kwargs["response"]
        if self.is_failure:
            return OrderResponse(status=FAILURE, error_msg="Unexpected error")
        else:
            order_id = uuid.uuid4().hex
            keys = VOrder.__fields__.keys()
            order_args = dict(order_id=order_id)
            is_user: bool = False
            userid: Optional[str] = None
            delay: int = self._delay
            for k, v in kwargs.items():
                if k == "userid":
                    userid = str(v).upper()
                    if userid in self.clients:
                        is_user = True
                if k == "delay":
                    delay = v
                elif k in keys:
                    order_args[k] = v
            try:
                resp = VOrder(**order_args)
                resp._delay = delay
                self._orders[order_args["order_id"]] = resp
                if is_user:
                    for user in self.users:
                        if user.userid == userid:
                            user.orders.append(resp)
                            break
                return OrderResponse(status=SUCCESS, data=resp)
            except ValidationError as e:
                errors: List = e.errors()
                num = len(errors)
                fld = errors[0].get("loc")[0]
                msg = errors[0].get("msg")
                error_msg = f"Found {num} validation errors; in field {fld} {msg}"
                return OrderResponse(status=FAILURE, error_msg=error_msg)

    def order_modify(
        self, order_id: str, **kwargs
    ) -> Union[OrderResponse, Dict[Any, Any]]:
        if "response" in kwargs:
            return kwargs["response"]
        if self.is_failure:
            return OrderResponse(status=FAILURE, error_msg="Unexpected error")
        if order_id not in self._orders:
            return OrderResponse(
                status=FAILURE,
                error_msg=f"Order id {order_id} not found on system",
            )
        attribs = ("price", "trigger_price", "quantity")
        dict(order_id=order_id)
        order = self.get(order_id)
        for attrib in attribs:
            if attrib in kwargs:
                setattr(order, attrib, kwargs[attrib])
        return OrderResponse(status=SUCCESS, data=order)

    def order_cancel(
        self, order_id: str, **kwargs
    ) -> Union[OrderResponse, Dict[Any, Any]]:
        if "response" in kwargs:
            return kwargs["response"]
        if self.is_failure:
            return OrderResponse(status=FAILURE, error_msg="Unexpected error")
        if order_id not in self._orders:
            return OrderResponse(
                status=FAILURE,
                error_msg=f"Order id {order_id} not found on system",
            )
        order = self.get(order_id)
        if order:
            if order.status == Status.COMPLETE:
                return OrderResponse(
                    status=FAILURE, error_msg=f"Order {order_id} already completed"
                )
            else:
                order.canceled_quantity = order.quantity - order.filled_quantity
                order.pending_quantity = 0
                return OrderResponse(status=SUCCESS, data=order)
        else:
            return OrderResponse(
                status=FAILURE,
                error_msg=f"Order id {order_id} not found on system",
            )

    def update_tickers(self, last_price: Dict[str, float]):
        """
        update tickers
        last_price
            dictionary of last traded price with key being
            the symbol name and value the last price
        """
        for k, v in last_price.items():
            ticker = self.tickers.get(k)
            if ticker:
                ticker.update(v)

    def _ltp(self, symbol: str) -> Optional[Dict[str, float]]:
        """
        get last traded price for a symbol
        """
        ticker = self.tickers.get(symbol)
        if ticker:
            return {symbol: ticker.ltp}
        else:
            return None

    def ltp(self, symbol: Union[str, Iterable]) -> Optional[Dict[str, float]]:
        """
        Get last traded prices for the given list of symbols
        """
        return _iterate_method(self._ltp, symbol)

    def _ohlc(self, symbol: str) -> Optional[Dict[str, OHLCV]]:
        ticker = self.tickers.get(symbol)
        if ticker:
            return {symbol: ticker.ohlc()}
        else:
            return None

    def ohlc(self, symbol: Union[str, Iterable]) -> Optional[Dict[str, OHLCV]]:
        """
        Get OHLC prices
        """
        return _iterate_method(self._ohlc, symbol)

    def _quote(self, symbol: str) -> Optional[Dict[str, VQuote]]:
        """
        return the quote for the symbol
        """
        ticker = self.tickers.get(symbol)
        if ticker:
            if ticker.orderbook:
                quote = VQuote(orderbook=ticker.orderbook, **ticker.ohlc().dict())
                return {symbol: quote}
            else:
                return None
        else:
            return None

    def quote(self, symbol: Union[str, Iterable]) -> Optional[Dict[str, VQuote]]:
        """
        return the quote for the symbol or list of symbols
        """
        return _iterate_method(self._quote, symbol)



from omspy.base import Broker, pre, post


class Paper(Broker):
    """
    An empty broker class
    """

    def __init__(self, orders=None, trades=None, positions=None):
        """
        Initialize the paper trader with some dummy data to return
        orders
            list of orders
        trades
            list of trades
        positions
            list of positions
        """
        self.__orders = orders
        self.__trades = trades
        self.__positions = positions
        super(Paper, self).__init__()

    def authenticate(self):
        return True

    @property
    @post
    def orders(self):
        return self.__orders if self.__orders else [{}]

    @property
    @post
    def trades(self):
        return self.__trades if self.__trades else [{}]

    @property
    @post
    def positions(self):
        return self.__positions if self.__positions else [{}]

    @pre
    def order_place(self, **kwargs):
        return kwargs

    @pre
    def order_modify(self, order_id: str, **kwargs):
        return kwargs.update({"order_id": order_id})

    @pre
    def order_cancel(self, order_id: str):
        return order_id
